
# This package exists to enable
# from degreedays import *
# to import everything that is likely to be needed by an API user in one go.
# It's a bit sloppy, but our classes generally have names that are likely to be
# unique, so it will probably be fine in most cases.  And it's great for example
# code.
# Oddly I couldn't get the example method described at
# http://docs.python.org/2/tutorial/modules.html
# of defining submodules in an __all__ to work.  But importing each package as
# below does work.
from degreedays.api import *
from degreedays.api.data import *
from degreedays.geo import *
from degreedays.time import *




from xml.sax import ContentHandler, parse
from degreedays.api import * #@UnusedWildImport
from degreedays.api.data import * #@UnusedWildImport
from degreedays.geo import * #@UnusedWildImport
from degreedays.time import * #@UnusedWildImport
import datetime

# Don't actually use this - just take a function.
class CollectionCallback(object):
    def collectionEnded(self, collected):
        pass
    
class DelegateLifetime(object):
    def onStart(self, qName, attributes):
        pass
    def isEnd(self, qName):
        pass
    
class SimpleDelegateLifetime(DelegateLifetime):
    def __init__(self, qName):
        self.__qName = qName
        self.__count = 1
    def isEnd(self, qName):
        if self.__qName == qName:
            self.__count -= 1
            if self.__count == 0:
                return True
        return False
    def onStart(self, qName, attributes):
        if self.__qName == qName:
            self.__count += 1
    
class CharacterCollectionStrategy(object):
    # Return none if it doesn't want the characters for this qName
    def getCollectionCallbackOrNone(self, qName, attributes):
        pass
    
class FixedCharacterCollectionStrategy(CharacterCollectionStrategy):
    def __init__(self, nameSet):
        self.__nameSet = nameSet
        self.__nameToCollectedStringMap = {}
    def getCollectionCallbackOrNone(self, qName, attributes):
        if qName in self.__nameSet:
            def callback(collected):
                self.__nameToCollectedStringMap[qName] = collected
            return callback
        else:
            return None
    def getCollectedOrNone(self, qName):
        return self.__nameToCollectedStringMap.get(qName)


class SimpleSaxHandler(ContentHandler):
    
    def __init__(self):
        ContentHandler.__init__(self)
        self.__delegateHandlingEndedCallback = None
        self.__parent = None
        self.__childDelegates = []
        self.__childDelegateLifetimes = []
        self.__currentDelegate = None
        self.__currentDelegateLifetime = None
        self.__characterCollectionStrategy = None
        # Following are non None if characters are being collected.  Then back
        # to None when we're done.
        self.__collectionCallbackOrNone = None
        self.__characterCollector = None
        
    def withDelegateHandlingEndedCallback(self, callback):
        self.__delegateHandlingEndedCallback = callback
        return self
    
    def setCharacterCollectionStrategy(self, strategy):
        self.__characterCollectionStrategy = strategy
        return strategy
    
    def __getRootAncestor(self):
        testHandler = self
        while testHandler.__parent is not None:
            testHandler = testHandler.__parent
        return testHandler
        
    def startElement(self, qName, attributes):
        if self.__currentDelegate is not None:
            self.__currentDelegateLifetime.onStart(qName, attributes)
            self.__currentDelegate.startElement(qName, attributes)
        else:
            if self.__characterCollectionStrategy is not None:
                callbackOrNone = self.__characterCollectionStrategy \
                    .getCollectionCallbackOrNone(qName, attributes)
                if callbackOrNone is not None:
                    self.collectCharacters(callbackOrNone)
            self.startElementImpl(qName, attributes)
     
    def startElementImpl(self, qName, attributes):
        pass
     
    def endElement(self, qName):
        isDelegated = False
        if self.__currentDelegate is not None:
            if self.__currentDelegateLifetime.isEnd(qName):
                self.__currentDelegate.delegateHandlingEnded()
                noParents = len(self.__childDelegates)
                if noParents > 0:
                    self.__currentDelegate = self.__childDelegates.pop()
                    self.__currentDelegateLifetime = \
                        self.__childDelegateLifetimes.pop() 
                else:
                    self.__currentDelegate = None
                    self.__currentDelegateLifetime = None
            else:
                self.__currentDelegate.endElement(qName)
                isDelegated = True
        if not isDelegated:
            if self.__collectionCallbackOrNone is not None:
                collected = ''.join(self.__characterCollector)
                self.__collectionCallbackOrNone(collected)
                # reset now to save memory.
                self.__characterCollector = None
                # Set to null so we know not to collect more.
                self.__collectionCallbackOrNone = None
            self.endElementImpl(qName);
    
    def endElementImpl(self, qName):
        pass
    
    def addDelegate(self, delegate, lifetimeOrQNameForEnd):
        delegate.__parent = self
        root = self.__getRootAncestor()
        if not isinstance(lifetimeOrQNameForEnd, DelegateLifetime):
            lifetimeOrQNameForEnd = \
                SimpleDelegateLifetime(lifetimeOrQNameForEnd)
        root.__addDelegateImpl(delegate, lifetimeOrQNameForEnd)
    
    def __addDelegateImpl(self, delegate, lifetime):
        if (self.__currentDelegate is not None):
            self.__childDelegates.append(self.__currentDelegate)
            self.__childDelegateLifetimes.append(self.__currentDelegateLifetime)
        self.__currentDelegate = delegate
        self.__currentDelegateLifetime = lifetime
    
    # Subclasses can override this if they want to be told when they are no
    # longer the delegate handler.  Override this to get the completed stuff
    # out of a delegate when it's done.
    def delegateHandlingEnded(self):
        if self.__delegateHandlingEndedCallback is not None:
            self.__delegateHandlingEndedCallback(self)
    
    # To be called by a subclass in {@link #startElementImpl}.  Probably 
    # unnecessary to call this if using a {@link CharacterCollectionStrategy}.
    def collectCharacters(self, collectionCallback):
        self.__characterCollector = []
        self.__collectionCallbackOrNone = collectionCallback;
    
    def characters(self, content):
        if self.__currentDelegate is not None:
            self.__currentDelegate.characters(content)
        elif self.__collectionCallbackOrNone is not None:
            self.__characterCollector.append(content)

REQUEST_UNITS_AVAILABLE = 'RequestUnitsAvailable'
MINUTES_TO_RESET = 'MinutesToReset'
# set and frozenset were both introduced in Python 2.4.  So there's no reason
# to use set when frozenset is better here (as it's immutable).
RATE_LIMIT_NAMES = frozenset((REQUEST_UNITS_AVAILABLE, MINUTES_TO_RESET))

class SaxHandlerForRateLimit(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.chars = self.setCharacterCollectionStrategy(
            FixedCharacterCollectionStrategy(RATE_LIMIT_NAMES))
    def getRateLimit(self):
        requestUnits = int(self.chars.getCollectedOrNone(
            REQUEST_UNITS_AVAILABLE))
        minutesToReset = int(self.chars.getCollectedOrNone(
            MINUTES_TO_RESET))
        return RateLimit(requestUnits, minutesToReset)
    
class SaxHandlerForResponseMetadata(SimpleSaxHandler):
    def startElementImpl(self, qName, attributes):
        if "RateLimit" == qName:
            def callback(handler):
                self.rateLimit = handler.getRateLimit()
            self.addDelegate(SaxHandlerForRateLimit() \
                .withDelegateHandlingEndedCallback(callback), qName)
    def getMetadata(self):
        return ResponseMetadata(self.rateLimit)

CODE = 'Code'
MESSAGE = 'Message'
FAILURE_NAMES = frozenset((CODE, MESSAGE))
    
class SaxHandlerForFailure(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.chars = self.setCharacterCollectionStrategy(
            FixedCharacterCollectionStrategy(FAILURE_NAMES))
    def getFailure(self):
        return Failure(
            self.chars.getCollectedOrNone(CODE),
            self.chars.getCollectedOrNone(MESSAGE))
        
class SaxHandlerForFailureResponse(SaxHandlerForFailure):
    def getResponse(self, metadata):
        return FailureResponse(metadata, self.getFailure())
    
def parseLongLat(attributes):
    return LongLat(float(attributes.getValue('longitude')),
        float(attributes.getValue('latitude')))

def parseMetresDistance(distanceString):
    return Distance.metres(float(distanceString))
   
# Could put this in time module, but not sure we want to commit to keeping it.
def parseIsoDate(dateString):
    split = dateString.split('-', 2)
    return datetime.date(int(split[0]), int(split[1]), int(split[2]))

def parsePercentageEstimated(peString):
    return float(peString)

def parsePercentageEstimatedAttribute(attributes):
    if 'pe' in attributes:
        return float(attributes.getValue('pe'))
    else:
        return 0.0

def parseDegreeDayValue(valueString):
    return float(valueString)

class SaxHandlerForTargetLocation(SimpleSaxHandler):
    def startElementImpl(self, qName, attributes):
        if 'LongLat' == qName:
            self.longLat = parseLongLat(attributes)

ID = 'Id'
ELEVATION_METRES = 'ElevationMetres'
DISPLAY_NAME = 'DisplayName'
STATION_NAMES = frozenset((ID, ELEVATION_METRES, DISPLAY_NAME))
class SaxHandlerForStation(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.chars = self.setCharacterCollectionStrategy(
            FixedCharacterCollectionStrategy(STATION_NAMES))
    def startElementImpl(self, qName, attributes):
        if 'LongLat' == qName:
            self.longLat = parseLongLat(attributes)
    def getStation(self):
        return Station(
            self.chars.getCollectedOrNone(ID),
            self.longLat,
            parseMetresDistance(
                self.chars.getCollectedOrNone(ELEVATION_METRES)),
            self.chars.getCollectedOrNone(DISPLAY_NAME))  

METRES_FROM_TARGET = 'MetresFromTarget'
SOURCE_NAMES = frozenset((METRES_FROM_TARGET,))
class SaxHandlerForSource(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.chars = self.setCharacterCollectionStrategy(
            FixedCharacterCollectionStrategy(SOURCE_NAMES))
    def startElementImpl(self, qName, attributes):
        if 'Station' == qName:
            def callback(handler):
                self.station = handler.getStation()
            self.addDelegate(SaxHandlerForStation() \
                .withDelegateHandlingEndedCallback(callback), qName)
    def getSource(self):
        return Source(self.station, parseMetresDistance(
            self.chars.getCollectedOrNone(METRES_FROM_TARGET)))

class SaxHandlerForSources(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.sources = []
    def startElementImpl(self, qName, attributes):
        if 'Source' == qName:
            def callback(handler):
                self.sources.append(handler.getSource())
            self.addDelegate(SaxHandlerForSource() \
                .withDelegateHandlingEndedCallback(callback), qName)
   
STATION_ID = "StationId"
LOCATION_DATA_RESPONSE_HEAD_NAMES = frozenset((STATION_ID,))
class SaxHandlerForLocationDataResponseHead(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.chars = self.setCharacterCollectionStrategy(
            FixedCharacterCollectionStrategy(LOCATION_DATA_RESPONSE_HEAD_NAMES))
    def startElementImpl(self, qName, attributes):
        if 'TargetLocation' == qName:
            def callback(handler):
                self.targetLongLat = handler.longLat
            self.addDelegate(SaxHandlerForTargetLocation() \
                .withDelegateHandlingEndedCallback(callback), qName)
        elif 'Sources' == qName:
            def callback(handler):
                self.sources = handler.sources
            self.addDelegate(SaxHandlerForSources() \
                .withDelegateHandlingEndedCallback(callback), qName)
    def fillDict(self, d):
        d['stationId'] = self.chars.getCollectedOrNone(STATION_ID)
        d['targetLongLat'] = self.targetLongLat
        d['sources'] = self.sources

PERCENTAGE_ESTIMATED = 'PercentageEstimated'
DATED_DATA_SET_HEAD_NAMES = frozenset((PERCENTAGE_ESTIMATED,))

class SaxHandlerForDatedDataSetHead(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.chars = self.setCharacterCollectionStrategy(
            FixedCharacterCollectionStrategy(DATED_DATA_SET_HEAD_NAMES))
    def fillDict(self, d):
        d['percentageEstimated'] = parsePercentageEstimated(
            self.chars.getCollectedOrNone(PERCENTAGE_ESTIMATED))

class SaxHandlerForDatedDataSetValues(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.setCharacterCollectionStrategy(self)
        self.datedDataValues = []
    def getCollectionCallbackOrNone(self, qName, attributes):
        if qName != 'V':
            return None
        firstDay = parseIsoDate(attributes.getValue('d'))
        if 'ld' in attributes:
            lastDay = parseIsoDate(attributes.getValue('ld'))
            self.dayRange = DayRange(firstDay, lastDay)
        else:
            self.dayRange = DayRange.singleDay(firstDay)
        self.percentageEstimated = parsePercentageEstimatedAttribute(attributes)
        return self.collectionEnded
    def collectionEnded(self, collected):
        value = parseDegreeDayValue(collected)
        datedDataValue = DatedDataValue(
            value, self.percentageEstimated, self.dayRange)
        self.datedDataValues.append(datedDataValue)
       
class SaxHandlerForDatedDataSet(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.params = {}
    def startElementImpl(self, qName, attributes):
        if 'Head' == qName:
            def callback(handler):
                handler.fillDict(self.params)
            self.addDelegate(SaxHandlerForDatedDataSetHead() \
                .withDelegateHandlingEndedCallback(callback), qName)
        elif 'Values' == qName:
            def callback(handler):
                self.params['values'] = handler.datedDataValues
            self.addDelegate(SaxHandlerForDatedDataSetValues() \
                .withDelegateHandlingEndedCallback(callback), qName)
    def getResult(self):
        return DatedDataSet(**self.params)

FIRST_YEAR = 'FirstYear'
LAST_YEAR = 'LastYear'
AVERAGE_DATA_SET_HEAD_NAMES = frozenset((FIRST_YEAR, LAST_YEAR))   
class SaxHandlerForAverageDataSetHead(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.chars = self.setCharacterCollectionStrategy(
            FixedCharacterCollectionStrategy(AVERAGE_DATA_SET_HEAD_NAMES))
    def fillDict(self, d):
        d['firstYear'] = int(self.chars.getCollectedOrNone(FIRST_YEAR))
        d['lastYear'] = int(self.chars.getCollectedOrNone(LAST_YEAR))
        
class SaxHandlerForAverageDataSetMonthlyValues(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.setCharacterCollectionStrategy(self)
        self.monthlyAverages = []
    def getCollectionCallbackOrNone(self, qName, attributes):
        if qName != 'M':
            return None
        self.percentageEstimated = parsePercentageEstimatedAttribute(attributes)
        return self.collectionEnded
    def collectionEnded(self, collected):
        value = parseDegreeDayValue(collected)
        dataValue = DataValue(value, self.percentageEstimated)
        self.monthlyAverages.append(dataValue)
        
class SaxHandlerForAverageDataSetValues(SimpleSaxHandler):
    def startElementImpl(self, qName, attributes):
        if 'Monthly' == qName:
            def callback(handler):
                self.monthlyAverages = handler.monthlyAverages
            self.addDelegate(SaxHandlerForAverageDataSetMonthlyValues() \
                .withDelegateHandlingEndedCallback(callback), qName)
        elif 'Annual' == qName:
            percentageEstimated = parsePercentageEstimatedAttribute(attributes)
            def collectionEnded(collected):
                value = parseDegreeDayValue(collected)
                self.annualAverage = DataValue(value, percentageEstimated)
            self.collectCharacters(collectionEnded)

class SaxHandlerForAverageDataSet(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.params = {}
    def startElementImpl(self, qName, attributes):
        if 'Head' == qName:
            def callback(handler):
                handler.fillDict(self.params)
            self.addDelegate(SaxHandlerForAverageDataSetHead() \
                .withDelegateHandlingEndedCallback(callback), qName)
        elif 'Values' == qName:
            def callback(handler):
                self.params['annualAverage'] = handler.annualAverage
                self.params['monthlyAverages'] = handler.monthlyAverages
            self.addDelegate(SaxHandlerForAverageDataSetValues() \
                .withDelegateHandlingEndedCallback(callback), qName)
    def getResult(self):
        return AverageDataSet(**self.params)

class SaxHandlerForDataSetFailure(SaxHandlerForFailure):
    def getResult(self):
        return self.getFailure()
        
class SaxHandlerForDataSets(SimpleSaxHandler):
    def __init__(self):
        SimpleSaxHandler.__init__(self)
        self.resultsMap = {}
    def __addHandler(self, qName, attributes, newHandler):
        def callback(handler):
            self.resultsMap[attributes.getValue('key')] = handler.getResult()
        newHandler = newHandler.withDelegateHandlingEndedCallback(callback)
        self.addDelegate(newHandler, qName)    
    def startElementImpl(self, qName, attributes):
        if 'DatedDataSet' == qName:
            self.__addHandler(qName, attributes, SaxHandlerForDatedDataSet())
        elif 'AverageDataSet' == qName:
            self.__addHandler(qName, attributes, SaxHandlerForAverageDataSet())
        elif 'Failure' == qName:
            self.__addHandler(qName, attributes, SaxHandlerForDataSetFailure())
    def getDataSets(self, dataSpecsOrNone):
        return DataSets(dataSpecsOrNone, self.resultsMap)
    
class SaxHandlerForLocationDataResponse(SimpleSaxHandler):
    def __init__(self, locationDataRequestOrNone):
        SimpleSaxHandler.__init__(self)
        if locationDataRequestOrNone is None:
            self.dataSpecsOrNone = None
        else:
            self.dataSpecsOrNone = locationDataRequestOrNone.dataSpecs
        self.params = {}
    def startElementImpl(self, qName, attributes):
        if 'Head' == qName:
            def callback(handler):
                handler.fillDict(self.params)
            self.addDelegate(SaxHandlerForLocationDataResponseHead() \
                .withDelegateHandlingEndedCallback(callback), qName)
        elif 'DataSets' == qName:
            def callback(handler):
                self.params['dataSets'] = handler.getDataSets(
                    self.dataSpecsOrNone)
            self.addDelegate(SaxHandlerForDataSets() \
                .withDelegateHandlingEndedCallback(callback), qName)
    def getResponse(self, metadata):
        self.params['metadata'] = metadata
        return LocationDataResponse(**self.params)

class SaxHandlerForResponse(SimpleSaxHandler):
    def __init__(self, requestOrNone):
        SimpleSaxHandler.__init__(self)
        self.requestOrNone = requestOrNone
    def startElementImpl(self, qName, attributes):
        if 'Metadata' == qName:
            def callback(handler):
                self.metadata = handler.getMetadata()
            self.addDelegate(SaxHandlerForResponseMetadata() \
                .withDelegateHandlingEndedCallback(callback), qName)
        elif 'LocationDataResponse' == qName:
            def callback(handler):
                self.response = handler.getResponse(self.metadata)
            self.addDelegate(
                SaxHandlerForLocationDataResponse(self.requestOrNone) \
                    .withDelegateHandlingEndedCallback(callback), qName)
        elif 'Failure' == qName:
            def callback(handler):
                self.response = handler.getResponse(self.metadata)
            self.addDelegate(SaxHandlerForFailureResponse() \
                .withDelegateHandlingEndedCallback(callback), qName)     
            
class DefaultResponseParser(object):
    def parseResponse(self, responseStream, request):
        handler = SaxHandlerForResponse(request)
        parse(responseStream, handler)
        return handler.response
        



import degreedays._private as private
import degreedays.time
import degreedays.geo
from degreedays._private import XmlElement
import datetime
import degreedays.api as api
import re

__all__ = ['Location', 'GeographicLocation',
    'Period',
    'Breakdown', 'DatedBreakdown', 'AverageBreakdown',
    'Calculation',
    'DataSpec', 'DatedDataSpec', 'AverageDataSpec', 'DataSpecs',
    'LocationDataRequest', 'DataApi',
    'TemperatureUnit', 'Temperature',
    'LocationError', 'SourceDataError',
    'DataValue', 'DatedDataValue',
    'DataSet', 'DatedDataSet', 'AverageDataSet', 'DataSets',
    'Station', 'Source',
    'LocationDataResponse']

_ABSTRACT_INIT_FACTORY = ('This is an abstract superclass.  To create an '
    'instance use the class factory methods, for example: ')
_ABSTRACT_INIT_RESPONSE = ('This is an abstract superclass.  Only the '
    'subclasses %s can be instantiated directly, though usually the API would '
    'create them automatically.')

_LOCATION_EXAMPLE = ("Location.stationId('KNYC'), "
    "Location.longLat(degreedays.geo.LongLat(-122.45361, 37.80544)), "
    "Location.postalCode('10036', 'US'), "
    "Location.postalCode('WC2N 5DN', 'GB')")
class Location(private.Immutable):
    __slots__ = ()
    def __init__(self):
        raise TypeError(_ABSTRACT_INIT_FACTORY + _LOCATION_EXAMPLE + '.')
    @classmethod
    def stationId(cls, stationId):
        # Having to import impl for class factory methods like this is ugly...
        # But we can't do it at the top because then that's a circular import
        # (as impl depends on this module).
        # This seems to be the only way to avoid cluttering this module up
        # conceptually for the user by forcing them to see way more classes
        # than they'll need.
        # Apparently at least this isn't a performance hit:
        # http://stackoverflow.com/questions/477096/python-import-coding-style
        # "Because of the way python caches modules, there isn't a performance
        # hit. In fact, since the module is in the local namespace, there is a
        # slight performance benefit to importing modules in a function."
        import degreedays.api.data.impl as impl
        return impl.StationIdLocation(stationId)
    @classmethod
    def longLat(cls, longLat):
        import degreedays.api.data.impl as impl
        return impl.LongLatLocation(longLat)
    @classmethod
    def postalCode(cls, postalCode, countryCode):
        import degreedays.api.data.impl as impl
        return impl.PostalCodeLocation(postalCode, countryCode)
    @classmethod
    def _check(cls, param, paramName='location'):
        if not isinstance(param, Location):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                Location, _LOCATION_EXAMPLE))
        return param

_GEOGRAPHIC_LOCATION_EXAMPLE = (
    "Location.longLat(degreedays.geo.LongLat(-122.45361, 37.80544)), "
    "Location.postalCode('10036', 'US'), "
    "Location.postalCode('WC2N 5DN', 'GB').")   
class GeographicLocation(Location):
    __slots__ = ()
    def __init__(self):
        raise TypeError(_ABSTRACT_INIT_FACTORY +
            _GEOGRAPHIC_LOCATION_EXAMPLE + '.')

_PERIOD_EXAMPLE = ("Period.latestValues('5'), "
    "Period.dayRange(degreedays.time.DayRange("
        "datetime.date(2012, 1, 1), "
        "datetime.date(2012, 10, 31)))")  
class Period(private.Immutable):
    __slots__ = ()
    def __init__(self):
        raise TypeError(_ABSTRACT_INIT_FACTORY + _PERIOD_EXAMPLE + '.')
    @classmethod
    def latestValues(cls, numberOfValues, minimumNumberOfValuesOrNone=None):
        import degreedays.api.data.impl as impl
        return impl.LatestValuesPeriod(
            numberOfValues, minimumNumberOfValuesOrNone)
    @classmethod
    def dayRange(cls, dayRange, minimumDayRangeOrNone=None):
        import degreedays.api.data.impl as impl
        return impl.DayRangePeriod(dayRange, minimumDayRangeOrNone)
    @classmethod
    def _check(cls, param, paramName='period'):
        if not isinstance(param, Period):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                Period, _PERIOD_EXAMPLE))
        return param

_BREAKDOWN_EXAMPLE = ('DatedBreakdown.daily(Period.latestValues(30)), '
    'DatedBreakdown.monthly(Period.latestValues(12)), '
    'AverageBreakdown.fullYears(Period.latestValues(5))')   
class Breakdown(private.Immutable):
    __slots__ = ()
    def __init__(self):
        raise TypeError(_ABSTRACT_INIT_FACTORY + _BREAKDOWN_EXAMPLE + '.')

_DATED_BREAKDOWN_EXAMPLE = ('DatedBreakdown.daily(Period.latestValues(30)), '
    'DatedBreakdown.monthly(Period.latestValues(12)), '
    'DatedBreakdown.yearly(Period.latestValues(5))')
class DatedBreakdown(Breakdown):
    __slots__ = ()
    def __init__(self):
        raise TypeError(_ABSTRACT_INIT_FACTORY + _DATED_BREAKDOWN_EXAMPLE + '.')
    @classmethod
    def daily(cls, period):
        import degreedays.api.data.impl as impl
        return impl.DailyBreakdown(period)
    @classmethod
    def weekly(cls, period, firstDayOfWeek):
        import degreedays.api.data.impl as impl
        return impl.WeeklyBreakdown(period, firstDayOfWeek)
    @classmethod
    def monthly(cls, period, startOfMonth=degreedays.time.StartOfMonth(1)): 
        import degreedays.api.data.impl as impl
        return impl.MonthlyBreakdown(period, startOfMonth)
    @classmethod
    def yearly(cls, period, startOfYear=degreedays.time.StartOfYear(1, 1)):
        import degreedays.api.data.impl as impl
        return impl.YearlyBreakdown(period, startOfYear)
    @classmethod
    def _check(cls, param, paramName='datedBreakdown'):
        if not isinstance(param, DatedBreakdown):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                DatedBreakdown, _DATED_BREAKDOWN_EXAMPLE))
        return param

_AVERAGE_BREAKDOWN_EXAMPLE = (
    'AverageBreakdown.fullYears(Period.latestValues(5)) for a ' +
    '5-year average covering the 5 most recent full calendar years') 
class AverageBreakdown(Breakdown):
    __slots__ = ()
    def __init__(self):
        raise TypeError(_ABSTRACT_INIT_FACTORY +
            _AVERAGE_BREAKDOWN_EXAMPLE, '.')
    @classmethod
    def fullYears(cls, period):
        import degreedays.api.data.impl as impl
        return impl.FullYearsAverageBreakdown(period)
    @classmethod
    def _check(cls, param, paramName='averageBreakdown'):
        if not isinstance(param, AverageBreakdown):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                AverageBreakdown, _AVERAGE_BREAKDOWN_EXAMPLE))
        return param

# TemperatureUnit follows the pattern shown by one of the answers at:
# http://stackoverflow.com/questions/36932/
# Though we define CELSIUS and FAHRENHEIT inside the class (initially as None,
# then set after the definition), so as to make them appear in PyDev
# intellisense.
# The metaclass stuff is to enable people to iterate over the units like:
# "for u in TemperatureUnit".  It's complicated to make it compatible with
# Python 2 and 3.  Our approach is based on that described at:
# http://mikewatkins.ca/2008/11/29/python-2-and-3-metaclasses/
# We have to pass a dict into TemperatureUnitMetaclass, which would imply
# mutability, but following the guidance at
# http://stackoverflow.com/questions/9654133/
# we pass a dictionary containing empty __slots__ in, which works great at
# ensuring immutability.
class _TemperatureUnitMetaclass(type):
    def __iter__(self):
        for u in (TemperatureUnit.CELSIUS, TemperatureUnit.FAHRENHEIT):
            yield u
_TemperatureUnitSuper = _TemperatureUnitMetaclass('_TemperatureUnitSuper',
    (private.Immutable,), {'__slots__': ()})
class TemperatureUnit(_TemperatureUnitSuper):
    __slots__ = ('_shortChar', '_name', '__nameUpper',
        '_minTimesTen', '_maxTimesTen')
    __map = {}
    CELSIUS = None # Set later
    FAHRENHEIT = None # as above
    def __init__(self, shortChar):
        self._shortChar = shortChar
        if shortChar == 'C' and TemperatureUnit.CELSIUS is None:
            self._name = 'Celsius'
            self._minTimesTen = -2730
            self._maxTimesTen = 30000
        elif shortChar == 'F' and TemperatureUnit.FAHRENHEIT is None:
            self._name = 'Fahrenheit'
            self._minTimesTen = -4594
            self._maxTimesTen = 54320
        else:
            raise TypeError('This is not built for direct '
                'instantiation.  Please use either TemperatureUnit.CELSIUS or '
                'TemperatureUnit.FAHRENHEIT.')
        self.__nameUpper = self._name.upper();
    def _equalityFields(self):
        return (self._shortChar,)
    def __str__(self):
        return self._name
    def __repr__(self):
        return 'TemperatureUnit.' + self.__nameUpper
    @classmethod
    def _check(cls, param, paramName='temperatureUnit'):
        if not isinstance(param, TemperatureUnit):
            raise TypeError(private.wrongTypeString(param, paramName,
                TemperatureUnit,
                'TemperatureUnit.CELSIUS or TemperatureUnit.FAHRENHEIT'))
        return param
    
TemperatureUnit.CELSIUS = TemperatureUnit('C')
TemperatureUnit.FAHRENHEIT = TemperatureUnit('F')

def _formatTemperature(value):
    s = '%.1f' % value
    # We did use rstrip('.0') but that didn't work with 0.0.
    if s.endswith('.0'):
        return s[:-2]
    return s

def _formatTemperatureTimesTen(timesTenValue):
    # Must divide by 10.0 to avoid it doing whole integer division (in python
    # 2 anyway; 3 is different: http://www.python.org/dev/peps/pep-0238/)
    return _formatTemperature(timesTenValue / 10.0)

_TEMPERATURE_EXAMPLE = ('Temperature.celsius(15.5), '
    'Temperature.fahrenheit(65)')  
class Temperature(private.Immutable):
    __slots__ = ('__valueTimesTen', '__unit')
    def __init__(self, value, unit):
        value = private.checkNumeric(value, 'value')
        # round returns a float, but int(...) of that should work fine,
        # according to:
        # http://stackoverflow.com/questions/3387655/
        self.__valueTimesTen = int(round(value * 10))
        if (self.__valueTimesTen < unit._minTimesTen or 
                self.__valueTimesTen > unit._maxTimesTen):
            raise ValueError('Invalid %s temperature %s - cannot be less ' 
                'than %s or greater than %s' %
                (unit.__name, _formatTemperature(value), 
                    _formatTemperatureTimesTen(unit._minTimesTen),
                    _formatTemperatureTimesTen(unit._maxTimesTen)))
        self.__unit = TemperatureUnit._check(unit, 'unit')
    def _equalityFields(self):
        return (self.__valueTimesTen, self.__unit)
    @classmethod
    def celsius(cls, value):
        return Temperature(value, TemperatureUnit.CELSIUS)
    @classmethod
    def fahrenheit(cls, value):
        return Temperature(value, TemperatureUnit.FAHRENHEIT)
    @property
    def value(self): 
        # Must divide by 10.0 for same reason as in _formatTemperatureTimesTen
        return self.__valueTimesTen / 10.0
    @property
    def unit(self): return self.__unit
    def __repr__(self):
        if self.__unit == TemperatureUnit.CELSIUS:
            return 'Temperature.celsius(' + _formatTemperatureTimesTen(
                self.__valueTimesTen) + ')'
        else:
            return 'Temperature.fahrenheit(' + _formatTemperatureTimesTen(
                self.__valueTimesTen) + ')'
    def __str__(self):
        return (_formatTemperatureTimesTen(self.__valueTimesTen) + ' ' +
            self.__unit._shortChar)
    def _toXml(self, elementNameExtra=''):
        return XmlElement(self.__unit._name + elementNameExtra).setValue(
            _formatTemperatureTimesTen(self.__valueTimesTen))
    @classmethod
    def _check(cls, param, paramName='temperature'):
        if not isinstance(param, Temperature):
            raise TypeError(private.wrongTypeString(param, paramName,
                Temperature, _TEMPERATURE_EXAMPLE))
        return param

_CALCULATION_EXAMPLE = (
    'Calculation.heatingDegreeDays(Temperature.celsius(15.5)), '
    'Calculation.coolingDegreeDays(Temperature.fahrenheit(65))')            
class Calculation(private.Immutable):
    __slots__ = ()
    def __init__(self):
        raise TypeError(_ABSTRACT_INIT_FACTORY + _CALCULATION_EXAMPLE + '.')
    @classmethod
    def heatingDegreeDays(cls, baseTemperature):
        import degreedays.api.data.impl as impl
        return impl.HeatingDegreeDaysCalculation(baseTemperature)
    @classmethod
    def coolingDegreeDays(cls, baseTemperature):
        import degreedays.api.data.impl as impl
        return impl.CoolingDegreeDaysCalculation(baseTemperature)
    @classmethod
    def _check(cls, param, paramName='calculation'):
        if not isinstance(param, Calculation):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                Calculation, _CALCULATION_EXAMPLE))
        return param

_DATA_SPEC_EXAMPLE = (
    'DataSpec.dated(calculation, datedBreakdown), ' +
    'DataSpec.average(calculation, averageBreakdown)')
class DataSpec(private.Immutable):
    __slots__ = ()
    def __init__(self):
        raise TypeError(_ABSTRACT_INIT_FACTORY + _DATA_SPEC_EXAMPLE + '.')
    @classmethod
    def dated(cls, calculation, datedBreakdown):
        return DatedDataSpec(calculation, datedBreakdown)
    @classmethod
    def average(cls, calculation, averageBreakdown):
        return AverageDataSpec(calculation, averageBreakdown)
    @classmethod
    def _check(cls, param, paramName='dataSpec'):
        if not isinstance(param, DataSpec):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                DataSpec, _DATA_SPEC_EXAMPLE))
        return param
    
class DatedDataSpec(DataSpec):
    __slots__ = ('__calculation', '__breakdown')
    def __init__(self, calculation, datedBreakdown):
        self.__calculation = Calculation._check(calculation)
        self.__breakdown = DatedBreakdown._check(datedBreakdown)
    def _equalityFields(self):
        return (self.__calculation, self.__breakdown)
    @property
    def calculation(self): return self.__calculation
    @property
    def breakdown(self): return self.__breakdown
    def __repr__(self):
        return 'DatedDataSpec(%r, %r)' % (self.__calculation, self.__breakdown)
    def _toXml(self):
        return XmlElement('DatedDataSpec') \
            .addChild(self.__calculation._toXml()) \
            .addChild(self.__breakdown._toXml())
            
class AverageDataSpec(DataSpec):
    __slots__ = ('__calculation', '__breakdown')
    def __init__(self, calculation, averageBreakdown):
        self.__calculation = Calculation._check(calculation)
        self.__breakdown = AverageBreakdown._check(averageBreakdown)
    def _equalityFields(self):
        return (self.__calculation, self.__breakdown)
    @property
    def calculation(self): return self.__calculation
    @property
    def breakdown(self): return self.__breakdown
    def __repr__(self):
        return ('AverageDataSpec(%r, %r)' %
            (self.__calculation, self.__breakdown))
    def _toXml(self):
        return XmlElement('AverageDataSpec') \
            .addChild(self.__calculation._toXml()) \
            .addChild(self.__breakdown._toXml())
            
_DATA_SPEC_KEY_REGEXP_STRING = '[-_.0-9a-zA-Z]{1,60}$'
_DATA_SPEC_KEY_REGEXP = re.compile(_DATA_SPEC_KEY_REGEXP_STRING)
# This assumes that it's already been checked to see that it's a string.
def _checkDataSpecStringKey(key):
    if not _DATA_SPEC_KEY_REGEXP.match(key):
        raise ValueError('Invalid DataSpec/DataSet key (%r) - it should match '
            'the regular expression %s.' % (key, _DATA_SPEC_KEY_REGEXP_STRING))
    return key    

_DATA_SPECS_EXAMPLE = ('DataSpecs(hddSpec), '
    'DataSpecs(hddSpec, cddSpec), '
    'DataSpecs(listOfDataSpecObjects)')            
class DataSpecs(private.Immutable):
    __slots__ = ('__specToKeyMap', '__keyToSpecMap', '__orderedKeys',
            '__specSet')
    def __init__(self, *args):
        self.__specToKeyMap = {}
        # Could use OrderedDict, but that is Python 2.7+ and I'm not sure we
        # want to restrict to that version just for that.
        self.__keyToSpecMap = {}
        self.__orderedKeys = []
        if len(args) == 1 and isinstance(args[0], dict):
            for key, item in private.getDictItemsIterable(args[0]):
                if not private.isString(key):
                    raise TypeError('If creating this with a dict (to specify '
                        'custom keys for the request XML), DataSpecs expects '
                        'string keys.  But the dict it was passed contained %r '
                        'as a key.' % key)
                elif not isinstance(item, DataSpec):
                    raise TypeError('If creating this with a dict (to specify '
                        'custom keys for the request XML), DataSpecs expects '
                        'the values associated with each key to be DataSpec '
                        'instances (e.g. DataSpec.dated(...) or '
                        'DataSpec.average(...)).  But for key %r it found %r' %
                        (key, item))
                key = _checkDataSpecStringKey(key)
                self.__specToKeyMap[item] = key
                self.__keyToSpecMap[key] = item
                self.__orderedKeys.append(key)
        else:
            def add(item):
                index = len(self.__orderedKeys)
                if index >= 100:
                    raise ValueError(
                        'Cannot have more than 100 DataSpec items.')
                if isinstance(item, DataSpec):
                    if not item in self.__specToKeyMap:
                        key = str(index) # simple integer key for now
                        self.__specToKeyMap[item] = key
                        self.__keyToSpecMap[key] = item
                        self.__orderedKeys.append(key)
                else:
                    # assume it's a sequence.
                    # Could do a try catch too, so can give a useful error
                    # message if it's not a sequence or a DataSpec.
                    for innerItem in item:
                        add(innerItem)
            for arg in args:
                add(arg)
        if len(self.__orderedKeys) == 0:
            raise ValueError('Must have at least one DataSpec')
        # We create specSet purely for a quick equality comparison, as
        # described at:
        # http://stackoverflow.com/questions/3210832/
        # Note it has to be frozenset, not set, because only frozenset (which
        # is immutable) is hashable.
        self.__specSet = frozenset(self.__specToKeyMap)
        # Make tuple so immutable.
        self.__orderedKeys = tuple(self.__orderedKeys)
    def _equalityFields(self):
        return (self.__specSet,)
    @property
    def keys(self): return self.__orderedKeys
    def __getitem__(self, key):
        return self.__keyToSpecMap[key]
    def getKey(self, dataSpec):
        return self.__specToKeyMap[dataSpec]
    def _dataSetsEquals(self, thisResults, thatDataSpecs, thatResultsMap):
        for spec, thisKey in private.getDictItemsIterable(self.__specToKeyMap):
            thisResult = thisResults[thisKey]
            thatKey = thatDataSpecs.__specToKeyMap[spec]
            thatResult = thatResultsMap[thatKey]
            if thisResult != thatResult:
                return False
        return True
    def _dataSetsHashCode(self, thisResultsMap):
        hashCode = 17
        # hashCode of the DataSpecs stored inside the DataSets.
        hashCode = 31 * hashCode + hash(self)
        # We want to sum the hashCodes of the DataSet objects corresponding to 
        # each unique DataSpec.
        # 
        # We need to sum without a multiplier because, if we used a multiplier,
        # then the result would depend on the order.  And we don't want that.
        # This is like how it's done in Java's AbstractSet.
        for key in private.getDictValuesIterable(self.__specToKeyMap):
            result = thisResultsMap[key]
            hashCode += hash(result)
        # Before python 2.2 (earlier than this library covers), Python int
        # operations used to result in an error if they got big or small enough
        # to overflow.  They never overflowed automatically like Java ints.
        # Before 3 they automatically turn into longs, after 3 then there is no
        # difference between ints and longs anyway - they're both the int type:
        # http://www.python.org/dev/peps/pep-0237/
        # http://stackoverflow.com/questions/4581842/
        # http://stackoverflow.com/questions/2104884/
        # The docs for __hash__ say it should return an integer, but that, since
        # 2.5 it can return a long, and the system will use the 32-bit hash of
        # that object.  See:
        # http://docs.python.org/2/reference/datamodel.html#object.__hash__
        # To make this work for Python 2.2 to 2.5, we check for int and return
        # the hash if it isn't one (because presumably it's a long).
        # We don't worry about pre-2.2 at all - that may throw an overflow
        # error above.
        if not isinstance(hashCode, int):
            hashCode = hash(hashCode)
        return hashCode
    def __repr__(self):
        # We make this include the keys.  It's more useful for DataSets.__repr__
        # and for debugging, if necessary.  Also the keys are pretty short.
        s = []
        for key in self.__orderedKeys:
            s.append("'%s': %r" % (key, self.__keyToSpecMap[key]))
        s.sort()
        return 'DataSpecs({' + ', '.join(s) + '})'
    def _toXml(self):
        parent = XmlElement('DataSpecs')
        for key in self.__orderedKeys:
            child = self.__keyToSpecMap[key]._toXml().addAttribute('key', key)
            parent.addChild(child)
        return parent
    @classmethod
    def _check(cls, param, paramName='dataSpecs'):
        if not isinstance(param, DataSpecs):
            raise TypeError(private.wrongTypeString(param, paramName,
                DataSpecs, _DATA_SPECS_EXAMPLE))
        return param

_LOCATION_DATA_REQUEST_EXAMPLE = (
    "LocationDataRequest(Location.stationId('EGLL'), "
        "DataSpecs(DataSpec.dated("
            "Calculation.heatingDegreeDays(Temperature.fahrenheit(65)), "
                "DatedBreakdown.monthly(Period.latestValues(12)))))")   
class LocationDataRequest(private.Immutable):
    __slots__ = ('__location', '__dataSpecs')
    def __init__(self, location, dataSpecs):
        self.__location = Location._check(location)
        self.__dataSpecs = DataSpecs._check(dataSpecs)
    def _equalityFields(self):
        return (self.__location, self.__dataSpecs)
    @property
    def location(self): return self.__location
    @property
    def dataSpecs(self): return self.__dataSpecs  
    def __repr__(self):
        return ('LocationDataRequest(%r, %r)' %
            (self.__location, self.__dataSpecs)) 
    def _toXml(self):
        return XmlElement('LocationDataRequest') \
            .addChild(self.__location._toXml()) \
            .addChild(self.__dataSpecs._toXml())
    @classmethod
    def _check(cls, param, paramName='locationDataRequest'):
        if not isinstance(param, LocationDataRequest):
            raise TypeError(private.wrongTypeString(param, paramName,
                LocationDataRequest, _LOCATION_DATA_REQUEST_EXAMPLE))
        return param
            
            
class LocationError(api.RequestFailureError):
    def __init__(self, failureResponse):
        api.RequestFailureError.__init__(self, failureResponse)
    @property
    def isDueToLocationNotRecognized(self):
        return self._testCode('LocationNotRecognized')
    @property
    def isDueToLocationNotSupported(self):
        return self._testCode('LocationNotSupported')

class SourceDataError(api.FailureError):
    def __init__(self, failure):
        api.FailureError.__init__(self, failure)
    @property
    def isDueToSourceDataErrors(self):
        return self._testCode('SourceDataErrors')
    @property
    def isDueToSourceDataCoverage(self):
        return self._testCode('SourceDataCoverage')

# return true if it's non-zero
def _checkPercentageEstimated(percentageEstimated):
    private.checkNumeric(percentageEstimated, 'percentageEstimated')
    if percentageEstimated > 0:
        if percentageEstimated > 100:
            raise ValueError('Invalid percentageEstimated %r: cannot be > 100.'
                % percentageEstimated)
        return True
    elif percentageEstimated < 0:
        raise ValueError('Invalid percentageEstimated %r: cannot be < 0')
    # It's zero.
    return False
        
class DataValue(private.Immutable):
    __slots__ = ()
    def __init__(self):
        # To be sure that __init__ isn't called on this. Our __new__ should mean
        # only subclasses have their __init__ called.
        raise TypeError()
    def __new__(cls, value, percentageEstimated):
        # Kind of based on example at:
        # http://stackoverflow.com/questions/5953759/
        if cls is DataValue:
            value = private.checkNumeric(value, 'value')
            if _checkPercentageEstimated(percentageEstimated):
                return _EstimatedDataValue(value, percentageEstimated)
            else:
                return _SimpleDataValue(value, 0.0)
        else:
            return private.Immutable.__new__(cls)
    @property
    def value(self): raise NotImplementedError()
    @property
    def percentageEstimated(self): raise NotImplementedError()

class _SimpleDataValue(DataValue):
    __slots__ = ('__value',)
    # Pass percentageEstimated to make the auto-call of __init__ that occurs
    # after the superclass __new__ method have the right number of arguments.
    # Essentially we need to accept the args that DataValue(...) was called
    # with and the easiest way to do this is to match the signature here (even
    # though we don't need percentageEstimated for this one as it wouldn't be
    # called unless it was zero).
    def __init__(self, value, percentageEstimated):
        self.__value = value
    @property
    def value(self): return self.__value
    @property
    def percentageEstimated(self): return 0.0
    def _equalityFields(self):
        return (self.__value,)
    def __str__(self):
        # %g works well for floats. %f adds unnecessary trailing zeroes.
        return '%g' % self.__value
    def __repr__(self):
        return 'DataValue(%g)' % self.value
    
class _EstimatedDataValue(DataValue):
    __slots__ = ('__value', '__percentageEstimated')
    # As for SimpleDataValue, it's best to match the signature of DataValue.
    def __init__(self, value, percentageEstimated):
        self.__value = value
        self.__percentageEstimated = percentageEstimated
    @property
    def value(self): return self.__value
    @property
    def percentageEstimated(self): return self.__percentageEstimated
    def _equalityFields(self):
        return (self.__value, self.__percentageEstimated)
    def __str__(self):
        return '%g (%g%% estimated)' % (self.value, self.percentageEstimated)
    def __repr__(self):
        return ('DataValue(%g, %g)' %
                (self.__value, self.__percentageEstimated))
        
class DatedDataValue(DataValue):
    __slots__ = ()
    def __init__(self):
        # Like for DataValue.
        raise TypeError()
    def __new__(cls, value, percentageEstimated, dayRange):
        if cls is DatedDataValue:
            value = private.checkNumeric(value, 'value')
            if dayRange.first == dayRange.last:
                if _checkPercentageEstimated(percentageEstimated):
                    return _EstimatedSingleDayDatedDataValue(
                        value, percentageEstimated, dayRange)
                else:
                    return _SingleDayDatedDataValue(value, 0.0, dayRange)
            else:
                if _checkPercentageEstimated(percentageEstimated):
                    return _EstimatedMultiDayDatedDataValue(
                        value, percentageEstimated, dayRange)
                else:
                    return _MultiDayDatedDataValue(value, 0.0, dayRange)
        else:
            # Need to go right up to the class above DataValue, as we don't want
            # to have to match DataValue's args.
            return private.Immutable.__new__(cls)
    @property
    def dayRange(self): raise NotImplementedError()
    @property
    def firstDay(self):
        return self.dayRange.first
    @property
    def lastDay(self):
        return self.dayRange.last
    def __repr__(self):
        return ('DatedDataValue(%g, %g, %r)' %
            (self.value, self.percentageEstimated, self.dayRange))
        
class _SingleDayDatedDataValue(DatedDataValue):
    __slots__ = ('__value', '__day')
    def __init__(self, value, percentageEstimated, dayRange):
        self.__value = value
        self.__day = dayRange.first
    def _equalityFields(self):
        return (self.__value, self.__day)
    @property
    def value(self): return self.__value
    @property
    def percentageEstimated(self): return 0.0
    @property
    def dayRange(self):
        return degreedays.time.DayRange(self.__day, self.__day)
    @property
    def firstDay(self): return self.__day
    @property
    def lastDay(self): return self.__day
    def __str__(self):
        return '%s: %g' % (self.__day, self.value)

class _EstimatedSingleDayDatedDataValue(DatedDataValue):
    __slots__ = ('__value', '__percentageEstimated', '__day')
    def __init__(self, value, percentageEstimated, dayRange):
        self.__value = value
        self.__percentageEstimated = percentageEstimated
        self.__day = dayRange.first
    def _equalityFields(self):
        return (self.__value, self.__percentageEstimated, self.__day)
    @property
    def value(self): return self.__value
    @property
    def percentageEstimated(self): return self.__percentageEstimated
    @property
    def dayRange(self):
        return degreedays.time.DayRange(self.__day, self.__day)
    @property
    def firstDay(self): return self.__day
    @property
    def lastDay(self): return self.__day
    def __str__(self):
        return ('%s: %g (%g%% estimated)' %
            (self.__day, self.__value, self.__percentageEstimated))

class _MultiDayDatedDataValue(DatedDataValue):
    __slots__ = ('__value', '__dayRange')
    def __init__(self, value, percentageEstimated, dayRange):
        self.__value = value
        self.__dayRange = dayRange
    def _equalityFields(self):
        return (self.__value, self.__dayRange)
    @property
    def value(self): return self.__value
    @property
    def percentageEstimated(self): return 0.0
    @property
    def dayRange(self): return self.__dayRange
    def __str__(self):
        return '%s: %g' % (self.__dayRange, self.value)

class _EstimatedMultiDayDatedDataValue(DatedDataValue):
    __slots__ = ('__value', '__percentageEstimated', '__dayRange')
    def __init__(self, value, percentageEstimated, dayRange):
        self.__value = value
        self.__percentageEstimated = percentageEstimated
        self.__dayRange = dayRange
    def _equalityFields(self):
        return (self.__value, self.__percentageEstimated, self.__dayRange)
    @property
    def value(self): return self.__value
    @property
    def percentageEstimated(self): return self.__percentageEstimated
    @property
    def dayRange(self): return self.__dayRange
    def __str(self):
        return ('%s: %g (%g%% estimated)' %
            (self.__dayRange, self.__value, self.__percentageEstimated))
            
class DataSet(private.Immutable):
    __slots__ = ()
    def __init__(self):
        raise TypeError(
            _ABSTRACT_INIT_RESPONSE % 'DatedDataSet and AverageDataSet')
    @property
    def percentageEstimated(self): raise NotImplementedError()
    @property
    def fullRange(self): raise NotImplementedError()
    
class DatedDataSet(DataSet):
    __slots__ = ('__percentageEstimated', '__values')
    def __init__(self, percentageEstimated, values):
        self.__percentageEstimated = percentageEstimated
        self.__values = tuple(values) # convert to immutable tuple.
    def _equalityFields(self):
        return (self.__percentageEstimated, self.__values)
    @property
    def percentageEstimated(self): return self.__percentageEstimated
    @property
    def values(self): return self.__values
    @property
    def fullRange(self):
        return degreedays.time.DayRange(self.__values[0].firstDay,
            self.__values[-1].lastDay)
    def __str__(self):
        s = []
        s.append('DatedDataSet(')
        s.append(str(len(self.__values)))
        s.append(' value')
        if len(self.__values) != 1:
            s.append('s')
        else:
            # include the value itself.  No need for % estimated as it will be
            # the same as for the whole set.
            s.append(' (%g)' % self.__values[0].value)
        s.append(' covering ')
        s.append(str(self.fullRange))
        if self.__percentageEstimated > 0:
            s.append(', %g%% estimated' % self.__percentageEstimated)
        s.append(')')
        return ''.join(s)
    def __repr__(self):
        return ('DatedDataSet(percentageEstimated=%g, values=%r)' %
            (self.__percentageEstimated, self.__values))
        
class AverageDataSet(DataSet):
    __slots__ = ('__firstYear', '__lastYear', '__annualAverage',
            '__monthlyAverages')
    def __init__(self, firstYear, lastYear, annualAverage, monthlyAverages):
        self.__firstYear = private.checkInt(firstYear, 'firstYear')
        self.__lastYear = private.checkInt(lastYear, 'lastYear')
        if firstYear > lastYear:
            raise ValueError('firstYear (%s) cannot be greater than lastYear '
                '(%s).' % (firstYear, lastYear))
        # TODO check types below
        self.__annualAverage = annualAverage
        self.__monthlyAverages = tuple(monthlyAverages) # make immutable
    def _equalityFields(self):
        return (self.__firstYear, self.__lastYear, self.__annualAverage,
            self.__monthlyAverages)
    @property
    def firstYear(self): return self.__firstYear
    @property
    def lastYear(self): return self.__lastYear
    @property
    def numberOfYears(self):
        return self.__lastYear - self.__firstYear + 1
    @property
    def annualAverage(self): return self.__annualAverage
    def monthlyAverage(self, monthIndexWithJanAs1):
        return self.__monthlyAverages[monthIndexWithJanAs1 - 1]
    @property
    def fullRange(self):
        return degreedays.time.DayRange(
            datetime.date(self.__firstYear, 1, 1),
            datetime.date(self.__lastYear, 12, 31))
    def __str__(self):
        return ('AverageDataSet(%d to %d, annualAverage = %s)' %
            (self.__firstYear, self.__lastYear, self.__annualAverage))
    def __repr__(self):
        return ('AverageDataSet(firstYear=%d, lastYear=%d, annualAverage=%r, '
            'monthlyAverages=%r)' %
            (self.__firstYear, self.__lastYear, self.__annualAverage,
                self.__monthlyAverages))
       
class DataSets(object):
    __slots__ = ('__dataSpecsOrNone', '__results', '__cachedHashCode')
    def __init__(self, dataSpecsOrNone, stringKeyToResultDict):
        self.__dataSpecsOrNone = dataSpecsOrNone
        # TODO - should probably validate below
        self.__results = stringKeyToResultDict.copy()
        self.__cachedHashCode = 0
    def __getitem__(self, dataSpecObjectOrStringKey):
        if private.isString(dataSpecObjectOrStringKey):
            # Unusual case, but useful for testing
            stringKey = dataSpecObjectOrStringKey
        elif isinstance(dataSpecObjectOrStringKey, DataSpec):
            if self.__dataSpecsOrNone is None:
                # http://stackoverflow.com/questions/1701199/ says ValueError
                # is the equivalent to IllegalStateException.
                raise ValueError("You can only use DataSpec objects to access "
                    "the corresponding DataSet objects if the DataSets object "
                    "has internal access to the original DataSpecs from the "
                    "request (which maps the string keys in the XML response "
                    "to the DataSpec objects used to request them). This "
                    "framework would usually ensure that the DataSets object "
                    "was created with access to the relevant DataSpecs object, "
                    "but are you perhaps parsing the XML response yourself? "
                    "If so, please ensure that your implementation passes the "
                    "DataSpecs object from the request into the DataSets "
                    "object, or do all your interrogation of this object using "
                    "string keys.")
            stringKey = self.__dataSpecsOrNone.getKey(dataSpecObjectOrStringKey)
        else:
            raise TypeError('Expecting a DataSpec object from the request '
                '(the usual way to use this), or a string key (less common, '
                'but occasionally useful for testing or debugging).  Instead, '
                'got %r' % dataSpecObjectOrStringKey)
        result = self.__results[stringKey]
        if isinstance(result, api.Failure):
            raise SourceDataError(result)
        return result
    def __eq__(self, other):
        if not isinstance(other, DataSets):
            return False
        if self.__dataSpecsOrNone is None:
            if other.__dataSpecsOrNone is not None:
                return False
            # need to use keys in comparison...  Which makes it simple.
            return self.__results == other.__results
        else:
            if self.__dataSpecsOrNone != other.__dataSpecsOrNone:
                return False
            return self.__dataSpecsOrNone._dataSetsEquals(self.__results,
                other.__dataSpecsOrNone, other.__results)
    def __ne__(self, other):
        return not (self == other) 
    def __hash__(self):
        # We cache the hash code, like in Java.  This should be thread safe,
        # like in Java, as reading or writing a single instance attribute is
        # supposed to be thread safe:
        # http://effbot.org/zone/thread-synchronization.htm
        # Assuming that's right, the worst that could happen with below is that
        # the hash code is calculated multiple times. But we should be confident
        # that this method will always return the same value.
        hashCode = self.__cachedHashCode
        if hashCode == 0:
            if self.__dataSpecsOrNone is None:
                # started out making a frozenset of the values and getting the
                # hash of that.  But actually using our own method should be
                # faster as we only need to iterate over the values instead of
                # copying them and so on.  Also a frozenset would remove
                # duplicates which would make hash collisions more likely (I 
                # think).
                hashCode = private.getDictValuesHash(self.__results)
            else:
                hashCode = self.__dataSpecsOrNone._dataSetsHashCode(
                    self.__results)
            self.__cachedHashCode = hashCode
        return hashCode
    def __repr__(self):
        if self.__dataSpecsOrNone is None:
            # There is no saved ordering of results (unless we added it, which
            # I suppose we could).
            return 'DataSets(%r, %r)' % (self.__dataSpecsOrNone, self.__results)
        # Otherwise use the ordering of the keys in DataSpecs.
        s = []
        for key in self.__dataSpecsOrNone.keys:
            s.append("'%s': %r" % (key, self.__results[key]))
        return 'DataSets(%r, {%s})' % (self.__dataSpecsOrNone, ', '.join(s))
    @classmethod
    def _check(cls, param, paramName='dataSets'):
        if not isinstance(param, DataSets):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                DataSets))
        return param
    
class Station(private.Immutable):
    __slots__ = ('__id', '__longLat', '__elevation', '__displayName')
    # Python has an id function, but popular opinion seems to be that it's fine
    # for an object to have an id attribute:
    # http://stackoverflow.com/questions/3497883/
    def __init__(self, id, longLat, elevation, displayName):
        self.__id = private.checkStationId(id, False)
        self.__longLat = degreedays.geo.LongLat._check(longLat)
        self.__elevation = degreedays.geo.Distance._check(
            elevation, 'elevation')
        self.__displayName = displayName
    def _equalityFields(self):
        return (self.__id, self.__longLat, self.__elevation, self.__displayName)
    def __repr__(self):
        return ("Station('%s', %r, %r, '%s')" %
            (self.__id, self.__longLat, self.__elevation, self.__displayName))
    @property
    def id(self): return self.__id
    @property
    def longLat(self): return self.__longLat
    @property
    def elevation(self): return self.__elevation
    @property
    def displayName(self): return self.__displayName
    @classmethod
    def _check(cls, param, paramName='station'):
        if not isinstance(param, Station):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                Station))
        return param
    
class Source(private.Immutable):
    __slots__ = ('__station', '__distanceFromTarget')
    def __init__(self, station, distanceFromTarget):
        self.__station = Station._check(station)
        self.__distanceFromTarget = degreedays.geo.Distance._check(
            distanceFromTarget, 'distanceFromTarget')
    def _equalityFields(self):
        return (self.__station, self.__distanceFromTarget)
    def __repr__(self):
        return 'Source(%r, %r)' % (self.__station, self.__distanceFromTarget)
    @property
    def station(self): return self.__station
    @property
    def distanceFromTarget(self): return self.__distanceFromTarget
    
class LocationDataResponse(api.Response):
    __slots__ = ('__metadata', '__stationId', '__targetLongLat', '__sources',
        '__dataSets')
    def __init__(self, metadata, stationId, targetLongLat, sources,
            dataSets):
        self.__metadata = metadata
        self.__stationId = private.checkStationId(stationId, False)
        self.__targetLongLat = degreedays.geo.LongLat._check(
            targetLongLat, 'targetLongLat')
        self.__sources = tuple(sources) # TODO check these
        self.__dataSets = DataSets._check(dataSets)
    def _equalityFields(self):
        # metadata isn't included in equality check.
        return (self.__stationId, self.__targetLongLat, self.__sources,
            self.__dataSets)
    def __repr__(self):
        return ("LocationDataResponse(%r, '%s', %r, %r, %r)" %
            (self.__metadata, self.__stationId, self.__targetLongLat,
                self.__sources, self.__dataSets))
    @property
    def metadata(self): return self.__metadata
    @property
    def stationId(self): return self.__stationId
    @property
    def targetLongLat(self): return self.__targetLongLat
    @property
    def sources(self): return self.__sources
    @property
    def dataSets(self): return self.__dataSets
    
    
class DataApi(object):
    def __init__(self, requestProcessor):
        self.__requestProcessor = requestProcessor
    def __checkAndGet(self, request, expectedResponseType):
        response = self.__requestProcessor.process(request);
        if isinstance(response, expectedResponseType):
            return response
        elif isinstance(response, api.FailureResponse):
            code = response.failure.code
            if code.startswith('Location'):
                raise LocationError(response)
            # for general exceptions
            raise api.RequestFailureError._create(response)
        else:
            raise ValueError('For a request of type %r, the RequestProcessor '
                'should return a response of type %r or a FailureResponse, not '
                '%r' % (type(request), expectedResponseType, type(response)))
    def getLocationData(self, locationDataRequest):
        LocationDataRequest._check(locationDataRequest)
        return self.__checkAndGet(
            locationDataRequest, LocationDataResponse)
    




from degreedays._private import Immutable, XmlElement
import degreedays._private as private

__all__ = ['LongLat', 'DistanceUnit', 'Distance']

# Allows for slight rounding errors on calculated longitude and latitude values.
_MARGIN = 0.001
class LongLat(Immutable):
    __slots__ = ('__longitude', '__latitude')
    def __fail(self, longitude, latitude):
        raise ValueError("Problem longitude (%r) or latitude (%r)." %
                         (longitude, latitude));
    def __init__(self, longitude, latitude):
        longitude = private.checkNumeric(longitude, 'longitude')
        latitude = private.checkNumeric(latitude, 'latitude')
        if longitude > 180.0:
            if longitude > 180.0 + _MARGIN:
                self.__fail(longitude, latitude)
            longitude = 180.0
        elif longitude < -180.0:
            if longitude < -180.0 - _MARGIN:
                self.__fail(longitude, latitude)
        if latitude > 90.0:
            if latitude > 90.0 + _MARGIN:
                self.__fail(longitude, latitude)
            latitude = 90.0
        elif latitude < -90.0:
            if latitude < -90.0 - _MARGIN:
                self.__fail(longitude, latitude)
            latitude = -90.0
        self.__longitude = longitude
        self.__latitude = latitude
    def _equalityFields(self):
        return (self.__longitude, self.__latitude)
    @property
    def longitude(self): return self.__longitude
    @property
    def latitude(self): return self.__latitude
    def __repr__(self):
        return 'LongLat(%f, %f)' % (self.__longitude, self.__latitude)
    def _toXml(self):
        # Will str localize it?  Hard to find in docs, but I tested it with
        # Windows set to a french locale and it didn't, which is good.
        return (XmlElement("LongLat")
                .addAttribute("longitude", str(self.__longitude))
                .addAttribute("latitude", str(self.__latitude)))
    @classmethod
    def _check(cls, param, paramName='longLat'):
        if not isinstance(param, LongLat):
            raise TypeError(private.wrongTypeString(param, paramName, LongLat,
                'LongLat(-74.0064, 40.7142)'))
        return param

# See TemperatureUnit for an explanation of this pattern
class _DistanceUnitMetaclass(type):
    def __iter__(self):
        for u in (DistanceUnit.METRES, DistanceUnit.KILOMETRES,
                DistanceUnit.FEET, DistanceUnit.MILES):
            yield u
_DistanceUnitSuper = _DistanceUnitMetaclass('_DistanceUnitSuper', (Immutable,),
    {'__slots__': ()})
_FEET_IN_METRE = 3.28083989501 
class DistanceUnit(_DistanceUnitSuper):
    __slots__ = ('__symbol', '__singularName', '__pluralName', '__isMetric',
        '__noOfBase', '__pluralNameUpper')
    __map = {}
    # Below are set later.  We want them defined here as class variables so that
    # they show up in intellisense if you type "DayOfWeek.".
    METRES = None
    KILOMETRES = None
    FEET = None
    MILES = None
    def __initImpl(self, singularName, pluralName, isMetric, noOfBase):
        self.__singularName = singularName
        self.__pluralName = pluralName
        self.__isMetric = isMetric
        self.__noOfBase = noOfBase
        self.__pluralNameUpper = pluralName.upper()
    def __init__(self, symbol):
        if symbol == 'm' and DistanceUnit.METRES is None:
            self.__initImpl('metre', 'metres', True, 1)
        elif symbol == 'km' and DistanceUnit.KILOMETRES is None:
            self.__initImpl('kilometre', 'kilometres', True, 1000)
        elif symbol == 'ft' and DistanceUnit.FEET is None:
            self.__initImpl('foot', 'feet', False, 1)
        elif symbol == 'mi' and DistanceUnit.MILES is None:
            self.__initImpl('mile', 'miles', False, 5280)
        else:
            raise TypeError('This is not built for direct '
                'instantiation.  Please use DistanceUnit.METRES, '
                'DistanceUnit.KILOMETRES, DistanceUnit.FEET, or '
                'DistanceUnit.MILES.')
        self.__symbol = symbol
        DistanceUnit.__map[symbol] = self
    def _equalityFields(self):
        return (self.__symbol,)
    def __convertValueToSameSchemeUnit(self, value, sameSchemeUnit):
        if self == sameSchemeUnit:
            return value
        elif sameSchemeUnit.__noOfBase == 1:
            return value * self.noOfBase
        elif self.__noOfBase == 1:
            return value / sameSchemeUnit.noOfBase
        else:
            raise ValueError('This should not happen with the range of units '
                'we have')
    def _convertValue(self, value, newUnit):
        if self.__isMetric == newUnit.__isMetric:
            return self.__convertValueToSameSchemeUnit(value, newUnit)
        elif self.__isMetric:
            inMetres = self.__convertValueToSameSchemeUnit(
                value, DistanceUnit.METRES)
            inFeet = inMetres * _FEET_IN_METRE
            return DistanceUnit.FEET.__convertValueToSameSchemeUnit(
                inFeet, newUnit)
        else:
            inFeet = self.__convertValueToSameSchemeUnit(
                value, DistanceUnit.FEET)
            inMetres = inFeet / _FEET_IN_METRE
            return DistanceUnit.METRES.__convertValueToSameSchemeUnit(
                inMetres, newUnit)
    def __str__(self):
        return self.__pluralName
    def __repr__(self):
        return 'DistanceUnit.' + self.__pluralNameUpper
    @property # as property to make it immutable (or as much as is possible)
    def _symbol(self): return self.__symbol
    @classmethod
    def _check(cls, param, paramName='distanceUnit'):
        if not isinstance(param, DistanceUnit):
            raise TypeError(private.wrongTypeString(param, paramName,
                DistanceUnit, 'DistanceUnit.METRES, DistanceUnit.KILOMETRES, '
                'DistanceUnit.FEET, DistanceUnit.MILES'))
        return param
DistanceUnit.METRES = DistanceUnit('m')
DistanceUnit.KILOMETRES = DistanceUnit('km')
DistanceUnit.FEET = DistanceUnit('ft')
DistanceUnit.MILES = DistanceUnit('mi')

class Distance(Immutable):
    __slots__ = ('__value', '__unit')
    def __init__(self, value, unit):
        self.__value = private.checkNumeric(value, 'value')
        self.__unit = DistanceUnit._check(unit, 'unit')
    def _equalityFields(self):
        return (self.__value, self.__unit)
    @property
    def value(self): return self.__value
    @property
    def unit(self): return self.__unit
    @classmethod
    def metres(cls, value):
        return Distance(value, DistanceUnit.METRES)
    @classmethod
    def kilometres(cls, value):
        return Distance(value, DistanceUnit.KILOMETRES)
    @classmethod
    def feet(cls, value):
        return Distance(value, DistanceUnit.FEET)
    @classmethod
    def miles(cls, value):
        return Distance(value, DistanceUnit.MILES)
    def __toUnit(self, newUnit):
        if newUnit == self.__unit:
            return self
        return Distance(self.__unit._convertValue(self.__value, newUnit),
            newUnit)
    def inMetres(self):
        return self.__toUnit(DistanceUnit.METRES)
    def inKilometres(self):
        return self.__toUnit(DistanceUnit.KILOMETRES)
    def inFeet(self):
        return self.__toUnit(DistanceUnit.FEET)
    def inMiles(self):
        return self.__toUnit(DistanceUnit.MILES)
    # "in" is a reserved word so we can't use that.  PEP8 recommends to add a
    # trailing underscore and this seems to be a kind of convention in Python.
    def in_(self, newUnit): 
        return self.__toUnit(newUnit)
    def __str__(self):
        return '%g %s' % (self.__value, self.__unit._symbol)
    def __repr__(self):
        return 'Distance.%s(%g)' % (self.__unit, self.__value)
    @classmethod
    def _check(cls, param, paramName='distance'):
        if not isinstance(param, Distance):
            raise TypeError(private.wrongTypeString(param, paramName, Distance,
                'Distance.metres(20)'))
        return param



import re

# For immutable types we started out using a variation on namedtuple, inspired
# by the code at:
# http://blog.thomnichols.org/2010/12/lightweight-data-types-in-python
# and also the source code for namedtuple:
# http://svn.python.org/view/python/branches/release26-maint/Lib/collections.py?revision=72955&view=markup
# Downside was that namedtuple required Python 2.6+, it wasn't so great for
# documentation (parameters specified as strings), it prohibited or at least
# made it difficult for us to have our own inheritance hierarchy, we couldn't
# have private fields in our immutable subclasses, we didn't have complete
# control over all methods, type-checking of constructor args involved
# implementing __new__ which was messy, and we had to override __slots__ in
# subclasses to prevent the overhead of a dict (we need to do that with a
# superclass too, but the point is that using this namedtuple thing didn't
# eliminate unnecessary code).
# So we shifted over to subclassing the following Immutable class instead. In
# commit 62 we got rid of all use of the original namedtuple approach, and
# removed that code from this module. 
#
# Good idea for subclasses to override __repr__ and possibly __str__ if they 
# want a shorter friendly string (by default __str__ will just use __repr__).
# Don't worry too much about making __repr__ return something that can be used
# to re-create the object using eval, but since it seems that is the ideal case,
# you might as well use that kind of object-creation format when possible.
# See notes at:
# http://stackoverflow.com/questions/1436703/
class Immutable(object):
    __slots__ = ()
    def _equalityFields(self):
        raise NotImplementedError()
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        else:
            return self._equalityFields() == other._equalityFields()
    def __ne__(self, other):
        return not (self == other)
    def __hash__(self):
        # For many of our subclasses we get less hash collisions if we include
        # the class type in the hash.  Otherwise, say,
        # DatedBreakdown.daily(Period.latestValues(1)) and
        # DatedBreakdown.monthly(Period.latestValues(1))
        # would have the same hash.
        # We include the class in the hash always for now.  We could of course
        # make it optional later if we wanted to.
        return hash((self.__class__,) + self._equalityFields())

# Rethrowing seems a bit tricky in Python 2.  I initially cobbled together
# something from 
# http://stackoverflow.com/questions/1350671/
# and http://nedbatchelder.com/blog/200711/rethrowing_exceptions_in_python.html
# and a bit of experimentation.
# But then I realized that what I had come up with wouldn't work in Python 3
# because the raise statement had changed - it no longer allowed multiple
# arguments.  So, for commit 66, I came up with this method instead, which works
# on Python 2.5, 2.7 and 3.3, and presumably others in between.
def raiseWrapped(wrapperExceptionClass):
    import sys
    info = sys.exc_info()
    # Note that info[1] is the exception itself.
    newException = wrapperExceptionClass(str(info[0]) + ': ' + str(info[1]))
    try:
        # Try the python 3 way of doing it first:
        newException.with_traceback(info[2])
    except:
        # Need to put below in exec otherwise it won't compile in Python 3.
        exec('raise newException, None, info[2]')
    # If it's got here, with_traceback worked so we just raise that.
    raise newException

# Work out once whether we have viewkeys and viewvalues, as although they say
# exceptions are cheap, they mean the try/except blocks are cheap...  If an
# exception is thrown, it's expensive:
# http://stackoverflow.com/questions/8107695/
def _runHasDictViews():
    testDict = {'key': 5}
    try:
        testDict.viewkeys()
        testDict.viewvalues()
        # It's Python 2, 2.7 or later
        return True
    except:
        # It's either Python 2 before 2.7, or python 3
        return False
_hasDictViews = _runHasDictViews()

def _runHasDictIterMethods():
    testDict = {'key': 5}
    try:
        testDict.iterkeys()
        testDict.itervalues()
        testDict.iteritems()
        # It's Python 2, 2.2 or later.
        return True
    except:
        # It's Python 2 before 2.2 or Python 3
        return False
_hasDictIterMethods = _runHasDictIterMethods()

# This will only work as expected if dictObject won't change.  Otherwise the
# view (python 2.7+ only) will change with it.  Note that viewkeys and
# viewvalues are gone in Python 3, where keys() and values() simply return a
# view instead.
def getSafeDictKeys(dictObject):
    if _hasDictViews:
        return dictObject.viewkeys()
    else:
        return dictObject.keys()
    
def getSafeDictValues(dictObject):
    if _hasDictViews:
        return dictObject.viewvalues()
    else:
        return dictObject.values()
    
def getDictValuesIterable(dictObject):
    if _hasDictIterMethods:
        # Python 2 (2.2 or later) for fast iterating over the values.
        return dictObject.itervalues()
    else:
        # Python 3 by default values returns an iterator.
        return dictObject.values()
    
def getDictItemsIterable(dictObject):
    if _hasDictIterMethods:
        return dictObject.iteritems()
    else:
        # In Python 3, items returns the same as iteritems used to.
        return dictObject.items()
 
# This must calculate the hash of the values in an order-independent way.   
def getDictValuesHash(dictObject, startingHashCode=0):
    hashCode = startingHashCode
    for v in getDictValuesIterable(dictObject):
        # This must be order-independent
        hashCode += hash(v)
    # Before python 2.2 (earlier than this library covers), Python int
    # operations used to result in an error if they got big or small enough
    # to overflow.  They never overflowed automatically like Java ints.
    # Before 3 they automatically turn into longs, after 3 then there is no
    # difference between ints and longs anyway - they're both the int type:
    # http://www.python.org/dev/peps/pep-0237/
    # http://stackoverflow.com/questions/4581842/
    # http://stackoverflow.com/questions/2104884/
    # The docs for __hash__ say it should return an integer, but that, since
    # 2.5 it can return a long, and the system will use the 32-bit hash of
    # that object.  See:
    # http://docs.python.org/2/reference/datamodel.html#object.__hash__
    # To make this work for Python 2.2 to 2.5, we check for int and return
    # the hash if it isn't one (because presumably it's a long).
    # We don't worry about pre-2.2 at all - that may throw an overflow
    # error above.
    if isinstance(hashCode, int):
        return hashCode
    else:
        return hash(hashCode)
    
class Attribute(Immutable):
    __slots__ = ('__name', '__value')
    def __init__(self, name, value):
        self.__name = name
        self.__value = value
    def _equalityFields(self):
        return (self.__name, self.__value)
    @property
    def name(self): return self.__name
    @property
    def value(self): return self.__value
    def appendString(self, s):
        s.append(self.__name)
        s.append("=\"")
        s.append(self.__value)
        s.append("\"")
    
class XmlElement(object):

    def __init__(self, qName):
        # NB below is how you define instance variables in python - if you put
        # them above they're class variables (i.e. like static in Java).
        self.qName = qName
        self.attributes = []
        # can contain XmlElement or string because you can add a string child.
        self.children = []
        self.value = None
    
    def addAttribute(self, name, value):
        self.attributes.append(Attribute(name, str(value)));
        return self;


    def _addChildImpl(self, child):
        if (self.value):
            raise ValueError(
                "Can't have a String value and child elements together")
        self.children.append(child)
        return self
   

    def addXmlStringAsChild(self, xmlString):
        self._addChildImpl(xmlString);
        
    # This returns this, not the child, as coding made it clear that that was
    # more useful.  Don't confuse it with newChild which returns the
    # child.
    def addChild(self, child):
        return self._addChildImpl(child)
        return self

    def newChild(self, qName):
        newChild = XmlElement(qName)
        return self.addChild(newChild)

    def setValue(self, value):
        # check if list is empty
        # http://stackoverflow.com/questions/53513/
        if (self.children):
            raise ValueError(
                    "Can't have child elements and a String value together")
        self.value = str(value)
        return self;

    def appendXmlString(self, stringList):
        stringList.append("<")
        stringList.append(self.qName)
        for a in self.attributes:
            stringList.append(" ")
            a.appendString(stringList)
        if (not self.children and not self.value):
            stringList.append("/>")
        else:
            stringList.append(">")
            if (self.value):
                stringList.append(self.value)
            else:
                for c in self.children:
                    if isinstance(c, XmlElement):
                        c.appendXmlString(stringList)
                    else:
                        stringList.append(c)
            stringList.append("</")
            stringList.append(self.qName)
            stringList.append(">")

    def toXmlString(self):
        # making a list and joining seems to be an efficient way to make a big
        # python string.
        # http://www.skymind.com/~ocrow/python_string/
        stringList = []
        self.appendXmlString(stringList)
        return ''.join(stringList)

# From http://stackoverflow.com/questions/11301138/
try:
    basestring  # attempt to evaluate basestring
    def isString(s):
        return isinstance(s, basestring)
except NameError:
    def isString(s):
        return isinstance(s, str)

def logSafe(string):
    return string # TODO

def __getExampleCode(exampleCodeOrNone):
    if exampleCodeOrNone is None:
        return ''
    else:
        return ' (e.g. ' + exampleCodeOrNone + ')'

def fullNameOfClass(cls):
    # auto-imported classes like str have module __builtin__.  But that's a bit
    # confusing in error messages.  Below is how we test for that, to exclude
    # it.  __module__ can also be None according to the docs.
    if cls.__module__ is None or cls.__module__ == str.__class__.__module__:
        return cls.__name__
    return cls.__module__ + '.' + cls.__name__

def wrongTypeString(param, paramName, expectedType, exampleCodeOrNone=None):
    # str of a type is like <type 'object'>
    return ('%s is an instance of %s, when it should be an instance of %s%s.' % 
        (paramName, fullNameOfClass(param.__class__), 
            fullNameOfClass(expectedType), __getExampleCode(exampleCodeOrNone)))

def wrongSupertypeString(param, paramName, expectedType, 
        exampleCodeOrNone=None):
    return ('%s is an instance of %s, when it should be a subclass of %s%s.' % 
        (paramName, fullNameOfClass(param.__class__), 
            fullNameOfClass(expectedType), __getExampleCode(exampleCodeOrNone)))

# Precompiling doesn't actually make all that much difference, but it is
# faster.
# We're using match to test against this, and match looks for a match at the
# beginning.  So we don't need a ^, but we do need a $ (so as not to match
# a valid string with invalid stuff after it).
_STATION_ID_REGEXP_STRING = '[a-zA-Z0-9_-]{1,60}$' 
_STATION_ID_REGEXP = re.compile(_STATION_ID_REGEXP_STRING)
def checkStationId(stationId, beTolerant):
    if not isString(stationId):
        # Refer to it as 'station ID' - that way this works for methods that
        # take stationId and/or id.
        raise TypeError('The station ID should be a str.')
    if beTolerant:
        stationId = stationId.strip()
    if not _STATION_ID_REGEXP.match(stationId):
        raise ValueError('Invalid station ID (%r) - it should match the '
            'regular expression %s.' %
                (logSafe(stationId), _STATION_ID_REGEXP_STRING))
    return stationId

try:
    intTypes = (int, long)
except:
    # Python 3, where there is no long type.
    intTypes = int
    
try:
    import numbers
    numericTypes = numbers.Number
except:
    # Python 2, before 2.6
    try:
        import decimal
        # Python 2.4 or later
        numericTypes = (int, long, float, decimal.Decimal)
    except:
        # Before Python 2.4
        numercTypes = (int, long, float)
        
try:
    from math import isnan, isinf # if this works we're on 2.6+
except:
    def isnan(number):
        return (number != number)
    # found this worked on Python 2.5 whilst float('inf') did not.
    # At http://stackoverflow.com/questions/2919754/ it says that float('inf')
    # is guaranteed to work on 2.6, but that's pretty much irrelevant since on
    # 2.6 we have math.isinf anyway.
    plusInf = 1e30000
    minusInf = -plusInf
    def isinf(number):
        return (number == plusInf or number == minusInf)

def checkNumeric(param, paramName):
    if not isinstance(param, numericTypes):
        raise TypeError('%s should be a numeric type, but it was of type %s.' %
                        (paramName, fullNameOfClass(param.__class__)))
    elif isnan(param):
        raise ValueError('%s cannot be NaN.' % paramName)
    elif isinf(param):
        raise ValueError('%s cannot be +inf or -inf.' % paramName)
    return param

# This is not tolerant of floats as that could mask bugs.
def checkInt(param, paramName):
    if not isinstance(param, intTypes):
        raise TypeError('%s should be an int, but it was of type %s.' %
            (paramName, fullNameOfClass(param.__class__)))
    return param

def checkString(param, paramName):
    if not isString(param):
        raise TypeError('%s should be a string, but it was of type %s.' %
            (paramName, fullNameOfClass(param.__class__)))
    



# Keep this whole package private for the moment.  Think more
# about the design of XmlHttpRequestProcessor before making it public.  For
# example, named parameters are kind of cool, but they don't help us make
# wrappers of existing objects, so in that sense we could rather do with the
# builder pattern.

from threading import Lock
from time import time
from datetime import datetime, timedelta, tzinfo
from degreedays.api import FailureResponse, TransportError
from degreedays._private import * #@UnusedWildImport
from degreedays.api._xmlparsing import DefaultResponseParser
import random

# urllib stuff is different in python 2 and 3.
# We use the approach recommended at http://python3porting.com/noconv.html
try:
    from urllib.request import urlopen, Request
    from urllib.parse import urlencode
except ImportError:
    from urllib2 import urlopen, Request
    from urllib import urlencode
    
import zlib

class HttpRequest(object):
    def __init__(self, url, params):
        self._url = url
        self._params = params.copy()
    @property
    def url(self):
        return self._url
    def paramNames(self):
        return getSafeDictKeys(self._params)
    def getParamValue(self, paramName):
        return self._params[paramName]
    def params(self):
        return self._params.copy()

class RequestSecurityInfo(Immutable):
    __slots__ = ('__endpoint', '__accountKey', '__utcTimestamp',
        '__timestampString', '__random')
    def __init__(self, endpoint, accountKey, utcTimestamp, dateTimeFormatter,
            random):
        self.__endpoint = endpoint
        self.__accountKey = accountKey
        self.__utcTimestamp = utcTimestamp
        self.__timestampString = dateTimeFormatter.format(utcTimestamp)
        self.__random = random
    def _equalityFields(self):
        return (self.__endpoint, self.__accountKey, self.__utcTimestamp,
            self.__timestampString, self.__random)
    @property
    def endpoint(self): return self.__endpoint
    @property
    def accountKey(self): return self.__accountKey
    @property
    def utcTimestamp(self): return self.__utcTimestamp
    @property
    def timestampString(self): return self.__timestampString
    @property
    def random(self): return self.__random
        
class EncodedString(Immutable):
    __slots__ = ('__encoded', '__encoding')
    def __init__(self, encoded, encoding):
        self.__encoded = encoded
        self.__encoding = encoding
    def _equalityFields(self):
        return (self.__encoded, self.__encoding)
    @property
    def encoded(self): return self.__encoded
    @property
    def encoding(self): return self.__encoding

class Signature(Immutable):
    __slots__ = ('__bytes', '__method')
    def __init__(self, bytes, method):
        self.__bytes = bytes
        self.__method = method
    def _equalityFields(self):
        return (self.__bytes, self.__method)
    @property
    def bytes(self): return self.__bytes
    @property
    def method(self): return self.__method
    
    
class DefaultEndpointGetter(object):
    def getEndpoint(self, request):
        return 'http://apiv1.degreedays.net/xml'
    

class DefaultUtcTimestampFactory(object):
    # returns a datetime rather than a time, as a datetime can have better
    # precision.
    def newUtcTimestamp(self):
        return datetime.utcnow()


# Straight from the docs at
# http://docs.python.org/2/library/datetime.html#datetime.tzinfo
# we need this to convert non-UTC times to UTC.
ZERO = timedelta(0)
HOUR = timedelta(hours=1)
# A UTC class.
class UTC(tzinfo):
    """UTC"""
    def utcoffset(self, dt):
        return ZERO
    def tzname(self, dt):
        return "UTC"
    def dst(self, dt):
        return ZERO
utc = UTC()
   

class DefaultDateTimeFormatter(object):
    def format(self, datetime):
        if datetime.tzinfo is not None:
            # It's not a default UTC datetime.
            datetime = datetime.astimezone(utc)
        # isoformat docs say, if datetime.utcoffset() does not return None,
        # it will append a 6-char string like +HH:MM.  We are cautious here in
        # case it appends +00:00 or (perhaps unlikely) -00:00.
        s = datetime.isoformat()
        if (s.endswith('+00:00') or s.endswith('-00:00')):
            s = s[:6] # remove the last 6 characters
        if not s.endswith('Z'):
            s += 'Z'
        return s 

    
class DefaultRandomFactory(object):
    def newRandom(self):
        # this is the method used by the python-oauth2 library, which is
        # supposedly well tested so presumably it works OK.
        return str(random.randint(0, 100000000))
    

class DefaultRequestToXml(object):
    def getXml(self, request):
        return request._toXml().toXmlString()
    

class DefaultXmlRequestWrapper(object):
    def getWrappedXml(self, requestXml, requestSecurityInfo):
        security = XmlElement("SecurityInfo")
        security.addChild(XmlElement("Endpoint").setValue(
                requestSecurityInfo.endpoint))
        security.addChild(XmlElement("AccountKey").setValue(
                str(requestSecurityInfo.accountKey)))
        security.addChild(XmlElement("Timestamp").setValue(
                requestSecurityInfo.timestampString))
        security.addChild(XmlElement("Random").setValue(
                requestSecurityInfo.random))
        envelope = XmlElement("RequestEnvelope")
        envelope.addChild(security)
        envelope.addXmlStringAsChild(requestXml)
        return envelope.toXmlString()
    
    
class DefaultStringToBytes(object):
    def getBytes(self, string):
        # Could do
        # bytearray(string, 'utf-8') which returns a mutable byte array, or
        # bytes(string, 'utf-8') which returns an immutable byte array (bytes
        # object), but I think that's Python 3+ only.
        # Not entirely sure what string.encode returns, but it is recommended
        # at
        # http://stackoverflow.com/questions/7585435/
        return string.encode('utf-8')
        

   
import hmac
import hashlib

class DefaultSigner(object):
    def __init__(self, securityKey):
        self._hmacTemplate = hmac.new(securityKey.getBytes(),
            digestmod=hashlib.sha256)
    # returns bytes
    def getSignature(self, bytesObject):
        # Docs for 2.7.3 say the msg should be a string:
        # http://docs.python.org/2/library/hmac.html
        # Similarly it says it returns a string, though one which may contain
        # null bytes.
        # docs for 3 say msg should be a bytes object and digest returns a bytes
        # object:
        # http://docs.python.org/3.2/library/hmac.html
        # It could be that bytes and strings are pretty much the same thing in
        # earlier python...  I'm not sure...
        hmacObject = self._hmacTemplate.copy()
        # Note instead of this template/copy stuff we could just do the below
        # line.  But the idea was to avoid storing the SecurityKey longer than
        # necessary (as maybe hmac object protects it somehow).
        #hmacObject = hmac.new(self._securityKey.getBytes(), bytesObject, 
        #    hashlib.sha256())
        hmacObject.update(bytesObject)
        return Signature(hmacObject.digest(), 'HmacSHA256')
    
    
import base64
    
class DefaultBytesToEncodedString(object):
    def getEncodedString(self, bytesObject):
        encodedByteString = base64.urlsafe_b64encode(bytesObject)
        encodedString = encodedByteString.decode('utf-8')
        # rstrip('=') below will strip any trailing equals characters (however
        # many there may be).
        encodedString = encodedString.rstrip('=')
        return EncodedString(encodedString, 'base64url')

_oldPython = False
try:    
    from io import RawIOBase
except:
    # probably Python 2.5.  Use a dummy RawIOBase to get the code below to
    # compile, but actually skip all the gzip stuff altogether.  We might be
    # able to get it working, but without RawIOBase or bytearray it looks like
    # it would be pretty tricky.  Perhaps the best we could do would be to
    # load the XML stream into a string in memory and parse that.  But I don't
    # think it's worth bothering with as it's pretty unlikely anyone is using
    # 2.5 anyway.
    _oldPython = True
    class RawIOBase(object):
        pass
    
# Wrapper around a stream containing gzip-compressed content that enables it to
# be read as if it was uncompressed.
class DecompressingStream(RawIOBase):
    def __init__(self, compressedStream, minReadBuffer):
        self._compressed = compressedStream
        self._decompressionObject = zlib.decompressobj(16 + zlib.MAX_WBITS)
        self._minReadBuffer = minReadBuffer
        super(RawIOBase, self).__init__()
    def close(self):
        return self._compressed.close()
    def closed(self):
        return self._compressed.close()
    def fileno(self):
        return self._compressed.fileno  
    def isatty(self):
        return self._compressed.isatty()
    def readable(self):
        return self._compressed.readable()
    # don't allow readline or readlines - just leave it to the superclass to
    # handle those (or probably throw an exception.   
    # can't seek, so leave seek, seekable, tell, and truncate to the superclass
    # (which doesn't allow those by default).
    # similarly for writable, writelines, and flush (only relevant for writable
    # streams).  And also for write (which is a method of BufferedIOBase instead
    # of its superclass IOBase).
    # Now I think we need to implement readinto.  I think RawIOBase handles
    # read, by calling readinto   
    def readinto(self, b):
        toFill = len(b)
        filled = 0
        while filled < toFill:
            tryToFillLength = toFill - filled
            data = None
            if self._decompressionObject.unconsumed_tail:
                data = self._decompressionObject.decompress(
                    self._decompressionObject.unconsumed_tail, tryToFillLength)
            else:
                readFromCompressed = max(tryToFillLength, self._minReadBuffer)
                compressed = self._compressed.read(readFromCompressed)
                data = self._decompressionObject.decompress(
                    compressed, tryToFillLength)
            lengthGot = len(data)
            if lengthGot == 0:
                # Watch out for this...  decompress can return an empty array if
                # we specified a max_length that was too short.  This is
                # particularly likely for the first call, as presumably it has
                # to get past the encoding table or whatever before it can
                # de-compress anything.  But you can also get the same problem
                # later in the stream if you try to decompress chunks that are
                # too small.  So this is why we use _minReadBuffer to ensure
                # that enough data is available for decompression.
                break
            self._hasDecompressed = True
            b[filled:filled + lengthGot] = data
            filled += lengthGot
        return filled
    
    
class DefaultHttpRequestDispatcher(object):
    def dispatch(self, httpRequest):
        try:
            # To handle the gzipped stream we follow the sample at
            # http://rationalpie.wordpress.com/2010/06/02/python-streaming-gzip-decompression/
            dataString = urlencode(httpRequest.params())
            # Python 3 requires the POST data to be bytes.  Should this use
            # StringToBytes?  Maybe, but then also maybe not because this
            # is a conversion that's necessary for the Python urllib library
            # (i.e. required by this particular HttpRequestDispatcher) rather
            # than being something done to satisfy our API specification.
            data = dataString.encode('utf-8')
            # Providing data makes it do a POST instead of a GET.
            request = Request(httpRequest.url, data)
            if not _oldPython:
                request.add_header('Accept-encoding', 'gzip')
            response = urlopen(request)
            isGZipped = response.headers.get(
                'content-encoding', '').find('gzip') >= 0
            if not isGZipped:
                # it's a stream-like object already, and it's decompressed, so
                # we just return it.
                return response
            else:
                return DecompressingStream(response, 1024)
        except Exception:
            raiseWrapped(TransportError)
        
        
class PrintingResponseParser(object):
    def parseResponse(self, responseStream, request):
        # Just a temporary implementation
        s = ''
        while True:
            # Using small reads just for testing.
            read = responseStream.read(6)
            if len(read) == 0:
                break
            else:
                s += read.decode("utf-8")
        print(s)


class RateLimitCache(object):
    _RATE_LIMIT_FAILURE_CACHE_SECONDS = 59.5
    def __init__(self):
        self._lock = Lock()
        self._cachedRateLimitFailureResponse = None
        self._cachedRateLimitFailureSetSeconds = 0

    def getCachedResponseOrNone(self):
        self._lock.acquire()
        try:
            if self._cachedRateLimitFailureResponse is not None:
                nowSeconds = time()
                if nowSeconds < self._cachedRateLimitFailureSetSeconds:
                    # The clock has changed - we don't know what's what and all
                    # we can reasonably do is clear the cached failure.
                    self._cachedRateLimitFailureResponse = None
                elif nowSeconds >= (self._cachedRateLimitFailureSetSeconds +
                        RateLimitCache._RATE_LIMIT_FAILURE_CACHE_SECONDS):
                    self._cachedRateLimitFailureResponse = None
                else:
                    return self._cachedRateLimitFailureResponse
            return None
        finally:
            self._lock.release()
    
    def onFreshResponse(self, response):
        # The check against requestUnitsAvailable is to save an exception
        # being generated unnecessarily each time there's a failure that
        # has nothing to do with the rate limit. Note also that something
        # might fail because of the rate limit, but there might still be
        # request units left if e.g. the request needed 10 request units,
        # but they only had 5. Wouldn't want to cache a response like that,
        # so we test for an available count of zero as well as a rate-limit
        # failure.
        if (isinstance(response, FailureResponse) and
                response.metadata.rateLimit.requestUnitsAvailable == 0):
            code = response.failure.code
            if (code.startswith('RateLimit') and
                    not code.startswith('RateLimitOnLocationChanges')):
                self._lock.acquire()
                try:
                    self._cachedRateLimitFailureResonse = response
                    self._cachedRateLimitFailureSetSeconds = time()
                finally:
                    self._lock.release()

class XmlHttpRequestProcessor(object):
    _lock = Lock() # for accessing the map below
    _rateLimitCacheMap = {}
    # If making this package public we'd want some way to stop this from being
    # called other than by the builder.  Maybe passing the builder in would
    # be the solution.  If we need to add another processing component, we 
    # don't really want optional parameters for those to retain compatibility,
    # as the builder would be neater.
    def __init__(self, accountKey, endpointGetter, utcTimestampFactory,
            dateTimeFormatter, randomFactory, requestToXml, xmlRequestWrapper,
            stringToBytes, signer, bytesToEncodedString, httpRequestDispatcher,
            responseParser):
        self._accountKey = accountKey
        self._endpointGetter = endpointGetter
        self._utcTimestampFactory = utcTimestampFactory
        self._dateTimeFormatter = dateTimeFormatter
        self._randomFactory = randomFactory
        self._requestToXml = requestToXml
        self._xmlRequestWrapper = xmlRequestWrapper
        self._stringToBytes = stringToBytes
        self._signer = signer
        self._bytesToEncodedString = bytesToEncodedString
        self._httpRequestDispatcher = httpRequestDispatcher
        self._responseParser = responseParser
        XmlHttpRequestProcessor._lock.acquire()
        try:
            self._rateLimitCache = \
                XmlHttpRequestProcessor._rateLimitCacheMap.get(accountKey)
            if self._rateLimitCache is None:
                self._rateLimitCache = RateLimitCache()
                XmlHttpRequestProcessor._rateLimitCacheMap[accountKey] = \
                        self._rateLimitCache
        finally:
            XmlHttpRequestProcessor._lock.release()
            
    def _getSecurityInfo(self, endpoint):
        timestamp = self._utcTimestampFactory.newUtcTimestamp()
        return RequestSecurityInfo(endpoint, self._accountKey,
            timestamp, self._dateTimeFormatter,
            self._randomFactory.newRandom())
            
    def _getHttpRequestData(self, request):
        endpoint = self._endpointGetter.getEndpoint(request)
        securityInfo = self._getSecurityInfo(endpoint)
        xml = self._requestToXml.getXml(request)
        wrappedXml = self._xmlRequestWrapper.getWrappedXml(xml, securityInfo)
        xmlBytes = self._stringToBytes.getBytes(wrappedXml)
        signature = self._signer.getSignature(xmlBytes)
        encodedRequest = self._bytesToEncodedString.getEncodedString(xmlBytes)
        # Note python has an immutable bytes type
        encodedSignature = self._bytesToEncodedString.getEncodedString(
            signature.bytes)
        params = {
            'request_encoding' : encodedRequest.encoding,
            'signature_method' : signature.method,
            'signature_encoding' : encodedSignature.encoding,
            'encoded_request' : encodedRequest.encoded,
            'encoded_signature' : encodedSignature.encoded
        }
        return HttpRequest(endpoint, params)
        
    def process(self, request):
        rateLimitCachedResponse = self._rateLimitCache.getCachedResponseOrNone()
        if rateLimitCachedResponse is not None:
            return rateLimitCachedResponse
        httpRequestData = self._getHttpRequestData(request)
        responseStream = self._httpRequestDispatcher.dispatch(httpRequestData)
        isSuccess = False
        try:
            response = None
            try:
                response = self._responseParser.parseResponse(
                    responseStream, request)
            except Exception:
                raiseWrapped(TransportError)
            self._rateLimitCache.onFreshResponse(response)
            isSuccess = True
            return response
        finally:
            if responseStream is not None:
                try:
                    responseStream.close();
                except Exception:
                    if isSuccess:
                        raiseWrapped(TransportError)
                    # otherwise just drop it, as we don't want it replacing the
                    # exception we've had already.
        

# Just a temporary method...  Really we'd want a builder I think.  Just not sure
# how best to do that in Python
def _buildXmlHttpRequestProcessor(accountKey, securityKey):
    return XmlHttpRequestProcessor(
        accountKey,
        DefaultEndpointGetter(),
        DefaultUtcTimestampFactory(),
        DefaultDateTimeFormatter(),
        DefaultRandomFactory(),
        DefaultRequestToXml(),
        DefaultXmlRequestWrapper(),
        DefaultStringToBytes(),
        DefaultSigner(securityKey),
        DefaultBytesToEncodedString(),
        DefaultHttpRequestDispatcher(),
        DefaultResponseParser())



from degreedays._private import XmlElement
import degreedays._private as private
import degreedays.time
import degreedays.geo
import degreedays.api.data as data
import re

# Define __all__ as it's the easiest way to make sure that none of our other
# imports (things from geo, time, and _private) become available to anyone that
# does an import * on this module.
__all__ = ['StationIdLocation', 'LongLatLocation', 'PostalCodeLocation',
    'LatestValuesPeriod', 'DayRangePeriod',
    'DailyBreakdown', 'MonthlyBreakdown', 'YearlyBreakdown',
    'FullYearsAverageBreakdown',
    'HeatingDegreeDaysCalculation', 'CoolingDegreeDaysCalculation']

class StationIdLocation(data.Location):
    __slots__ = ('__stationId',)
    def __init__(self, stationId):
        self.__stationId = private.checkStationId(stationId, True).upper()
    def _equalityFields(self):
        return (self.__stationId,)
    @property
    def stationId(self): return self.__stationId
    def __repr__(self):
        return "StationIdLocation('%s')" % self.__stationId
    def _toXml(self):
        return XmlElement("StationIdLocation").addChild(
                XmlElement("StationId").setValue(self.__stationId))  

class LongLatLocation(data.GeographicLocation):
    __slots__ = ('__longLat',)
    def __init__(self, longLat):
        self.__longLat = degreedays.geo.LongLat._check(longLat)
    def _equalityFields(self):
        return (self.__longLat,)
    @property
    def longLat(self): return self.__longLat
    def __repr__(self):
        return 'LongLatLocation(%r)' % self.__longLat
    def _toXml(self):
        return XmlElement("LongLatLocation").addChild(
                self.__longLat._toXml())     

# See notes in private on regexp testing.
_POSTAL_CODE_REGEXP_STRING = '[- 0-9a-zA-Z]{1,16}$' 
_POSTAL_CODE_REGEXP = re.compile(_POSTAL_CODE_REGEXP_STRING)
def _getValidPostalCode(postalCode):
    private.checkString(postalCode, 'postalCode')
    # Work on assumption that this could be user input, so auto-correct for
    # leading or trailing whitespace.
    postalCode = postalCode.strip()
    if not _POSTAL_CODE_REGEXP.match(postalCode):
        raise ValueError('Invalid postalCode (%r) - it should match the '
            'regular expression %s.' % (postalCode, _POSTAL_CODE_REGEXP_STRING))
    return postalCode.upper()

_COUNTRY_CODE_REGEXP = re.compile('[A-Z]{2}$')   
def _getValidCountryCode(countryCode):
    private.checkString(countryCode, 'countryCode')
    # Be strict about this - it's unlikely to be user input as the user would
    # probably select from a list of countries mapped to valid codes.  i.e. we
    # don't do any whitespace stripping, and we're strict about case.
    if not _COUNTRY_CODE_REGEXP.match(countryCode):
        raise ValueError('Invalid country code (%r) - it should be 2 upper-'
            'case alphabetical characters (A - Z).' % countryCode)
    return countryCode
    
class PostalCodeLocation(data.GeographicLocation):
    __slots__ = ('__postalCode', '__countryCode')
    def __init__(self, postalCode, countryCode):
        self.__postalCode = _getValidPostalCode(postalCode)
        self.__countryCode = _getValidCountryCode(countryCode)
    def _equalityFields(self):
        return (self.__postalCode, self.__countryCode)
    @property
    def postalCode(self): return self.__postalCode
    @property
    def countryCode(self): return self.__countryCode
    def __repr__(self):
        return ("PostalCodeLocation('%s', '%s')" %
            (self.__postalCode, self.__countryCode))
    def _toXml(self):
        e = XmlElement("PostalCodeLocation")
        e.addChild(XmlElement("PostalCode").setValue(self.__postalCode))
        e.addChild(XmlElement("CountryCode").setValue(self.__countryCode))
        return e
                
class LatestValuesPeriod(data.Period):
    __slots__ = ('__numberOfValues', '__minimumNumberOfValuesOrNone')
    def __init__(self, numberOfValues, minimumNumberOfValuesOrNone = None):
        self.__numberOfValues = private.checkInt(
            numberOfValues, 'numberOfValues')
        if numberOfValues < 1:
            raise ValueError('Invalid numberOfValues (%s) - cannot be < 1.'
                % numberOfValues)
        self.__minimumNumberOfValuesOrNone = minimumNumberOfValuesOrNone
        if minimumNumberOfValuesOrNone is not None:
            private.checkInt(minimumNumberOfValuesOrNone, 'minimumNumberOfValuesOrNone')
            if minimumNumberOfValuesOrNone < 1:
                raise ValueError('Invalid minimumNumberOfValuesOrNone (%s) - cannot '
                    'be < 1 (though it can be None if there is to be no '
                    'minimum (the default).' % minimumNumberOfValuesOrNone)
            elif minimumNumberOfValuesOrNone > numberOfValues:
                raise ValueError('Invalid minimumNumberOfValuesOrNone (%s) - cannot '
                    'be > numberOfValues (%s).' %
                    (minimumNumberOfValuesOrNone, numberOfValues))
    def _equalityFields(self):
        return (self.__numberOfValues, self.__minimumNumberOfValuesOrNone)
    @property
    def numberOfValues(self): return self.__numberOfValues
    @property
    def minimumNumberOfValuesOrNone(self): return self.__minimumNumberOfValuesOrNone
    def __repr__(self):
        if self.__minimumNumberOfValuesOrNone is None:
            return 'LatestValuesPeriod(%d)' % self.__numberOfValues
        else:
            return 'LatestValuesPeriod(%d, %d)' % \
                (self.__numberOfValues, self.__minimumNumberOfValuesOrNone)   
    def _toXml(self):
        e = XmlElement("LatestValuesPeriod").addChild(
            XmlElement("NumberOfValues").setValue(self.__numberOfValues))
        if (self.__minimumNumberOfValuesOrNone is not None):
            e.addChild(XmlElement("MinimumNumberOfValues").setValue(
                self.__minimumNumberOfValuesOrNone))
        return e
        
class DayRangePeriod(data.Period):
    __slots__ = ('__dayRange', '__minimumDayRangeOrNone')
    def __init__(self, dayRange, minimumDayRangeOrNone = None):
        self.__dayRange = degreedays.time.DayRange._check(dayRange)
        if minimumDayRangeOrNone is not None:
            degreedays.time.DayRange._check(minimumDayRangeOrNone)
        self.__minimumDayRangeOrNone = minimumDayRangeOrNone
    def _equalityFields(self):
        return (self.__dayRange, self.__minimumDayRangeOrNone)
    @property
    def dayRange(self): return self.__dayRange
    @property
    def minimumDayRangeOrNone(self): return self.__minimumDayRangeOrNone
    def __repr__(self):
        if self.__minimumDayRangeOrNone is None:
            return 'DayRangePeriod(%r)' % self.__dayRange
        else:
            return 'DayRangePeriod(%r, %r)' % \
                (self.__dayRange, self.__minimumDayRangeOrNone)
    def _toXml(self):
        e = XmlElement("DayRangePeriod").addChild(self.__dayRange._toXml())
        if (self.__minimumDayRangeOrNone is not None):
            e.addChild(self.__minimumDayRangeOrNone._toXml('MinimumDayRange'))
        return e;

class DailyBreakdown(data.DatedBreakdown):
    __slots__ = ('__period')
    def __init__(self, period):
        self.__period = data.Period._check(period)
    def _equalityFields(self):
        return (self.__period,)
    @property
    def period(self): return self.__period
    def __repr__(self):
        return 'DailyBreakdown(%r)' % self.__period
    def _toXml(self):
        return XmlElement("DailyBreakdown").addChild(self.__period._toXml())
    
class WeeklyBreakdown(data.DatedBreakdown):
    __slots__ = ('__period', '__firstDayOfWeek')
    def __init__(self, period, firstDayOfWeek):
        self.__period = data.Period._check(period)
        self.__firstDayOfWeek = degreedays.time.DayOfWeek._check(
            firstDayOfWeek, 'firstDayOfWeek')
    def _equalityFields(self):
        return (self.__period, self.__firstDayOfWeek)
    @property
    def period(self): return self.__period
    @property
    def firstDayOfWeek(self): return self.__firstDayOfWeek
    def __repr__(self):
        return ('WeeklyBreakdown(%r, %r)' %
            (self.__period, self.__firstDayOfWeek))
    def _toXml(self):
        e = XmlElement('WeeklyBreakdown') \
            .addChild(self.__period._toXml())
        e.addAttribute('firstDayOfWeek', str(self.__firstDayOfWeek))
        return e
        
class MonthlyBreakdown(data.DatedBreakdown):
    __slots__ = ('__period', '__startOfMonth')
    def __init__(self, period, startOfMonth=degreedays.time.StartOfMonth(1)):
        self.__period = data.Period._check(period)
        self.__startOfMonth = degreedays.time.StartOfMonth._check(startOfMonth)
    def _equalityFields(self):
        return (self.__period, self.__startOfMonth)
    @property
    def period(self): return self.__period
    @property
    def startOfMonth(self): return self.__startOfMonth
    def __repr__(self):
        if self.__startOfMonth.dayOfMonth == 1:
            return 'MonthlyBreakdown(%r)' % self.__period
        else:
            return ('MonthlyBreakdown(%r, %r)' % 
                (self.__period, self.__startOfMonth))
    def _toXml(self):
        e = XmlElement('MonthlyBreakdown') \
            .addChild(self.__period._toXml())
        if (self.__startOfMonth.dayOfMonth != 1):
            e.addAttribute('startOfMonth', str(self.__startOfMonth))
        return e;
        
class YearlyBreakdown(data.DatedBreakdown):
    __slots__ = ('__period', '__startOfYear')
    def __init__(self, period, startOfYear=degreedays.time.StartOfYear(1, 1)):
        self.__period = data.Period._check(period)
        self.__startOfYear = degreedays.time.StartOfYear._check(startOfYear)
    def _equalityFields(self):
        return (self.__period, self.__startOfYear)
    @property
    def period(self): return self.__period
    @property
    def startOfYear(self): return self.__startOfYear
    def __repr__(self):
        if self.__startOfYear == degreedays.time.StartOfYear(1, 1):
            return 'YearlyBreakdown(%r)' % self.__period
        else:
            return ('YearlyBreakdown(%r, %r)' %
                (self.__period, self.__startOfYear))
    def _toXml(self):
        e = XmlElement('YearlyBreakdown').addChild(self.period._toXml())
        if (self.startOfYear != degreedays.time.StartOfYear(1, 1)):
            e.addAttribute('startOfYear', str(self.startOfYear))
        return e;
    
class FullYearsAverageBreakdown(data.AverageBreakdown):
    __slots__ = ('__period',)
    def __init__(self, period):
        self.__period = data.Period._check(period)
    def _equalityFields(self):
        return (self.__period,)
    @property
    def period(self): return self.__period
    def __repr__(self):
        return 'FullYearsAverageBreakdown(%r)' % self.__period
    def _toXml(self):
        return (XmlElement('FullYearsAverageBreakdown')
            .addChild(self.period._toXml()))
            
class HeatingDegreeDaysCalculation(data.Calculation):
    __slots__ = ('__baseTemperature',)
    def __init__(self, baseTemperature):
        self.__baseTemperature = data.Temperature._check(
            baseTemperature, 'baseTemperature')
    def _equalityFields(self):
        return (self.__baseTemperature,)
    @property
    def baseTemperature(self): return self.__baseTemperature
    def __repr__(self):
        return 'HeatingDegreeDaysCalculation(%r)' % self.__baseTemperature
    def _toXml(self):
        return XmlElement('HeatingDegreeDaysCalculation').addChild(
                self.__baseTemperature._toXml('BaseTemperature'))

class CoolingDegreeDaysCalculation(data.Calculation):
    __slots__ = ('__baseTemperature',)
    def __init__(self, baseTemperature):
        self.__baseTemperature = data.Temperature._check(
            baseTemperature, 'baseTemperature')
    def _equalityFields(self):
        return (self.__baseTemperature,)
    @property
    def baseTemperature(self): return self.__baseTemperature
    def __repr__(self):
        return 'CoolingDegreeDaysCalculation(%r)' % self.__baseTemperature
    def _toXml(self):
        return XmlElement('CoolingDegreeDaysCalculation').addChild(
                self.__baseTemperature._toXml('BaseTemperature'))
    



from degreedays._private import Immutable
import degreedays._private as private
import string

__all__ = ('AccountKey', 'SecurityKey',
    'Request', 'Response',
    'ResponseMetadata',
    'RateLimit',
    'Failure',
    'FailureResponse',
    'DegreeDaysApiError', 'TransportError', 'FailureError',
    'RequestFailureError', 'InvalidRequestError', 'RateLimitError',
    'ServiceError',
    'DegreeDaysApi')

# lower-case letters apart from i, l, o, and digits apart from 0 and 1.  Doesn't
# include the hyphen which is allowed in certain places.
_KEY_CHARS = frozenset('abcdefghjkmnpqrstuvwxyz23456789')
_UPPER_CASE_KEY_LETTERS = frozenset('ABCDEFGHJKMNPQRSTUVWXYZ')
class _KeyValidator(object):
    def __init__(self, noBlocks, keyName):
        expectedNoHyphens = noBlocks - 1
        self.minLength = (noBlocks * 4)
        self.expectedLength = self.minLength + expectedNoHyphens
        self.noBlocks = noBlocks
        self.keyName = keyName
    def _getValidCharOrNone(self, c):
        if c in _KEY_CHARS:
            return c
        elif c in _UPPER_CASE_KEY_LETTERS:
            return c.lower()
        else:
            return None
    def getValidOrNone(self, s):
        if len(s) < self.minLength:
            return None
        validChars = []
        for c in s:
            validChar = self._getValidCharOrNone(c)
            if validChar is not None:
                length = len(validChars)
                if length >= self.expectedLength:
                    # Too many valid chars means it's invalid
                    return None
                if length > 0 and (length + 1) % 5 == 0:
                    validChars.append('-')
                validChars.append(validChar)
        if len(validChars) != self.expectedLength:
            return None
        return ''.join(validChars)
    def getValidOrThrow(self, s):
        valid = self.getValidOrNone(s)
        if valid is not None:
            return valid
        raise ValueError("Invalid %s - it should be a %d-character string made "
            "up of %d 4-character blocks separated by hyphens.  The 4-character"
            " blocks can contain lower-case Latin alphabet characters apart "
            "from 'i', 'l', and 'o', and numeric digits apart from '0' and "
            "'1'." % (self.keyName, self.expectedLength, self.noBlocks))

_ACCOUNT_KEY_VALIDATOR = _KeyValidator(3, 'stringAccountKey')
class AccountKey(Immutable):
    __slots__ = ('__stringAccountKey')
    def __init__(self, stringAccountKey):
        self.__stringAccountKey = _ACCOUNT_KEY_VALIDATOR.getValidOrThrow(
            stringAccountKey)
    def _equalityFields(self):
        return (self.__stringAccountKey,)
    def __repr__(self):
        return 'AccountKey(%s)' % self.__stringAccountKey
    def __str__(self):
        return self.__stringAccountKey
    @classmethod
    def _check(cls, param, paramName='accountKey'):
        if not isinstance(param, AccountKey):
            raise TypeError(private.wrongTypeString(param, paramName,
                AccountKey, "AccountKey('test-test-test')"))
        return param
    

# TODO - look into security of strings and bytes in Python, and the way that key
# is expected, and think about whether you really want to be storing or
# converting to bytes here.
_SECURITY_KEY_VALIDATOR = _KeyValidator(13, 'stringSecurityKey')
_SECURITY_KEY_EXAMPLE = "SecurityKey('" + ('test-' * 13).rstrip('-') + "')"
class SecurityKey(Immutable):
    __slots__ = ('__bytes')
    def __init__(self, stringSecurityKey):
        stringSecurityKey = _SECURITY_KEY_VALIDATOR.getValidOrThrow(
            stringSecurityKey)
        self.__bytes = stringSecurityKey.encode('utf-8')
    def _equalityFields(self):
        return (self.__bytes,)
    def __repr__(self):
        return 'SecurityKey(string not included for security reasons)'
    def getBytes(self):
        return self.__bytes
    def toStringKey(self):
        return self.__bytes.decode('utf-8')
    @classmethod
    def _check(cls, param, paramName='securityKey'):
        if not isinstance(param, SecurityKey):
            raise TypeError(private.wrongTypeString(param, paramName,
                AccountKey, _SECURITY_KEY_EXAMPLE))
        return param
    
class Request(Immutable):
    __slots__ = ()
    def __init__(self):
        raise TypeError('This is an abstract superclass.  To '
            'create an instance, try creating a LocationDataRequest instead ('
            'LocationDataRequest being a subclass of this).')
    
class RateLimit(Immutable):
    __slots__ = ('__requestUnitsAvailable', '__minutesToReset')
    def __init__(self, requestUnitsAvailable, minutesToReset):
        self.__requestUnitsAvailable = private.checkInt(
            requestUnitsAvailable, 'requestUnitsAvailable')
        self.__minutesToReset = private.checkInt(
            minutesToReset, 'minutesToReset')
        if requestUnitsAvailable < 0:
            raise ValueError('Invalid requestUnitsAvailable (%s) - cannot be '
                '< 0.' % requestUnitsAvailable)
        if minutesToReset < 1:
            raise ValueError('Invalid minutesToReset (%s) - cannot be < 1.' %
                minutesToReset)
    def _equalityFields(self):
        return (self.__requestUnitsAvailable, self.__minutesToReset)
    def __repr__(self):
        return ('RateLimit(requestUnitsAvailable=%d, minutesToReset=%d)' %
            (self.__requestUnitsAvailable, self.__minutesToReset))
    @property
    def requestUnitsAvailable(self): return self.__requestUnitsAvailable
    @property
    def minutesToReset(self): return self.__minutesToReset
    @classmethod
    def _check(cls, param, paramName='rateLimit'):
        if not isinstance(param, RateLimit):
            raise TypeError(private.wrongTypeString(param, paramName,
                RateLimit))
        return param
    
class ResponseMetadata(Immutable):
    __slots__ = ('__rateLimit')
    def __init__(self, rateLimit):
        self.__rateLimit = RateLimit._check(rateLimit)
    def _equalityFields(self):
        return (self.__rateLimit)
    def __repr__(self):
        return 'ResponseMetadata(%r)' % self.__rateLimit
    @property
    def rateLimit(self): return self.__rateLimit
    @classmethod
    def _check(cls, param, paramName='responseMetadata'):
        if not isinstance(param, ResponseMetadata):
            raise TypeError(private.wrongTypeString(param, paramName,
                ResponseMetadata))
        return param
    
class Failure(Immutable):
    __slots__ = ('__code', '__message')
    def __init__(self, code, message):
        self.__code = code
        self.__message = message
    def _equalityFields(self):
        return (self.__code, self.__message)
    def __repr__(self):
        return "Failure('%s', '%s')" % (self.__code, self.__message)
    @property
    def code(self): return self.__code
    @property
    def message(self): return self.__message
    @classmethod
    def _check(cls, param, paramName='failure'):
        if not isinstance(param, Failure):
            raise TypeError(private.wrongTypeString(param, paramName, Failure))
        return param
    
class Response(Immutable):
    __slots__ = ()
    def __init__(self):
        raise TypeError('This is an abstract superclass.  '
            'LocationDataResponse is a subclass that can be instantiated '
            'directly (though typically you would submit a LocationDataRequest '
            'and let the API create the LocationDataResponse automatically).')
    @property
    def metadata(self): raise NotImplementedError()

class FailureResponse(Response):
    __slots__ = ('__metadata', '__failure')
    def __init__(self, metadata, failure):
        self.__metadata = ResponseMetadata._check(metadata, 'metadata')
        self.__failure = Failure._check(failure)
    def _equalityFields(self):
        # metadata is not considered in equality.
        return (self.__failure,)
    def __repr__(self):
        return 'FailureResponse(%r, %r)' % (self.__metadata, self.__failure)
    @property
    def metadata(self): return self.__metadata
    @property
    def failure(self): return self.__failure
    @classmethod
    def _check(cls, param, paramName='failureResponse'):
        if not isinstance(param, FailureResponse):
            raise TypeError(private.wrongTypeString(param, paramName, 
                FailureResponse))
        return param
    
class DegreeDaysApiError(Exception):
    pass

class TransportError(DegreeDaysApiError):
    pass

class FailureError(DegreeDaysApiError):
    def __init__(self, failure):
        self.__failure = Failure._check(failure)
    @property
    def failure(self): return self.__failure
    def _testCode(self, code):
        return self.__failure.code.startswith(code)
    def __str__(self):
        # TODO - check this is working as you'd want if you print the exception
        # or whatever.  Maybe you need the class name in here too?
        return str(self.__failure)

class RequestFailureError(FailureError):
    def __init__(self, failureResponse):
        FailureResponse._check(failureResponse)
        FailureError.__init__(self, failureResponse.failure)
        self.__responseMetadata = failureResponse.metadata
    @property
    def responseMetadata(self):
        return self.__responseMetadata
    @classmethod
    def _create(cls, failureResponse):
        code = failureResponse.failure.code
        if code.startswith('InvalidRequest'):
            return InvalidRequestError(failureResponse);
        elif code.startswith('RateLimit'):
            return RateLimitError(failureResponse);
        elif code.startsWith('Service'):
            return ServiceError(failureResponse);
        else:
            # Shouldn't happen, unless new codes are added in.
            return RequestFailureError(failureResponse)

class InvalidRequestError(RequestFailureError):
    def __init__(self, failureResponse):
        RequestFailureError.__init__(self, failureResponse)
    @property
    def isDueToInvalidRequestAccount(self):
        return self._testCode('InvalidRequestAccount')
    @property
    def isDueToInvalidRequestSignature(self):
        return self._testCode('InvalidRequestSignature')
    @property
    def isDueToInvalidRequestTimestamp(self):
        return self._testCode('InvalidRequestTimestamp')

class RateLimitError(RequestFailureError):
    def __init__(self, failureResponse):
        RequestFailureError.__init__(self, failureResponse)
    @property
    def isDueToRateLimitOnLocationChanges(self):
        return self._testCode('RateLimitOnLocationChanges')

class ServiceError(RequestFailureError):
    def __init__(self, failureResponse):
        RequestFailureError.__init__(self, failureResponse)
    @property
    def isDueToServiceTemporarilyDown(self):
        return self._testCode('ServiceTemporarilyDown')
    @property
    def isDueToServiceUnexpectedError(self):
        return self._testCode('ServiceUnexpectedError') 

class DegreeDaysApi(object):
    def __init__(self, requestProcessor):
        self.__requestProcessor = requestProcessor
        
    @classmethod
    def fromKeys(cls, accountKey, securityKey):
        AccountKey._check(accountKey)
        SecurityKey._check(securityKey)
        # This import must come below the stuff above like FailureResponse and
        # TransportException that _processing imports itself.  Otherwise I guess
        # it makes some sort of circular import issue.  Really this module
        # shouldn't depend on processing, but I think it's OK for the
        # convenience that this method brings.
        from degreedays.api._processing import _buildXmlHttpRequestProcessor
        return DegreeDaysApi(_buildXmlHttpRequestProcessor(
            accountKey, securityKey))
        
    @property
    def dataApi(self):
        from degreedays.api.data import DataApi
        return DataApi(self.__requestProcessor)



from distutils.core import setup

def _readme():
    with open('README.txt') as f:
        return f.read()

setup(
    name='degreedays',
    version='0.9.1',
    description='Degree Days.net API Python Client Library',
    long_description=_readme(),
    author='BizEE Software',
    author_email='info@bizeesoftware.com',
    url='http://www.degreedays.net/api/python',
    # The classifiers list doesn't have Apache 2.0 listed specifically, so we
    # spell it out here.
    license='Apache License, Version 2.0',
    # setup finds modules automatically, but you must list the packages.
    packages=['degreedays', 'degreedays.api', 'degreedays.api.data'],
    classifiers=[ # From http://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering'
    ]
)




#from datetime import date

from degreedays._private import Immutable
from degreedays._private import XmlElement
import degreedays._private as private

__all__ = ['DayRange', 'DayOfWeek', 'StartOfMonth', 'StartOfYear']

_DAY_RANGE_EXAMPLE = ('DayRange(datetime.date(2012, 1, 1), '
        'datetime.date(2012, 11, 30)) '
    'or DayRange.singleDay(datetime.date.today() - datetime.timedelta(1)) '
    '(for yesterday only)')
# using datetime.date is perfect for us as it's timezone independent.
class DayRange(Immutable):
    __slots__ = ('__first', '__last')
    def __init__(self, first, last):
        if (first > last):
            raise ValueError("first (%r) cannot be > last (%r)" % (first, last))
        self.__first = first
        self.__last = last
    def _equalityFields(self):
        return (self.__first, self.__last)
    @classmethod
    def singleDay(cls, firstAndLast):
        return DayRange(firstAndLast, firstAndLast)
    @property
    def first(self): return self.__first
    @property
    def last(self): return self.__last
    def __len__(self):
        return (self.last - self.first).days
    def __repr__(self):
        return 'DayRange(%r, %r)' % (self.__first, self.__last)
    def _toXml(self, elementName = "DayRange"):
        # NB \ is line-continuation character, as mentioned at:
        # http://stackoverflow.com/questions/8683178/
        return XmlElement(elementName) \
                .addAttribute("first", self.__first.isoformat()) \
                .addAttribute("last", self.__last.isoformat())
    @classmethod
    def _check(cls, param, paramName='dayRange'):
        if not isinstance(param, DayRange):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                DayRange, _DAY_RANGE_EXAMPLE))
        return param

_DAY_OF_WEEK_EXAMPLE = ('DayOfWeek.MONDAY, DayOfWeek.TUESDAY, ..., or '
    'DayOfWeek(0) for Monday etc. (using the int constants that Python uses in '
    'its calendar module)')
# This metaclass stuff is to make it possible to iterate over the values, like
# "for dow in DayOfWeek".  See TemperatureUnit for an explanation and
# references.
class _DayOfWeekMetaclass(type):
    def __iter__(self):
        for i in range(7):
            yield DayOfWeek(i)
_DayOfWeekSuper = _DayOfWeekMetaclass('_DayOfWeekSuper', (Immutable,),
    {'__slots__': ()})            
class DayOfWeek(_DayOfWeekSuper):
    __slots__ = ('__index', '__name', '__nameUpper', '__isoIndex')
    __map = {}
    # Below are set later.  We want them defined here as class variables so that
    # they show up in intellisense if you type "DayOfWeek.".
    MONDAY = None
    TUESDAY = None
    WEDNESDAY = None
    THURSDAY = None
    FRIDAY = None
    SATURDAY = None
    SUNDAY = None
    def __init__(self, index):
        if index == 0:
            self.__name = 'Monday'
        elif index == 1:
            self.__name = 'Tuesday'
        elif index == 2:
            self.__name = 'Wednesday'
        elif index == 3:
            self.__name = 'Thursday'
        elif index == 4:
            self.__name = 'Friday'
        elif index == 5:
            self.__name = 'Saturday'
        elif index == 6:
            self.__name = 'Sunday'
        else:
            raise ValueError('Invalid int value for day of week (%r) - '
                'expecting int between 0 (Monday) and 6 (Sunday).')
        self.__nameUpper = self.__name.upper()
        self.__index = index
        self.__isoIndex = index + 1
    # We need this because of the way we're caching values and allowing direct
    # use of the constructor with indexes e.g. DayOfWeek(0) for Monday.  This
    # prevents direct use of the constructor from creating a new instance each
    # time.
    def __new__(cls, index):
        existing = DayOfWeek.__map.get(index, None)
        if existing is not None:
            return existing
        newItem = super(DayOfWeek, cls).__new__(cls)
        # We don't need to call __init__... python does it for us after this
        # __new__ method has returned the un-initialized item.  And it passes
        # the index parameter to __init__ as well.
        DayOfWeek.__map[index] = newItem
        return newItem
    def _equalityFields(self):
        return (self.__index,)
    @property
    def index(self): return self.__index
    @property
    def isoIndex(self): return self.__isoIndex
    def __str__(self):
        return self.__name
    def __repr__(self):
        return 'DayOfWeek.' + self.__nameUpper
    @classmethod
    def _check(cls, param, paramName='dayOfWeek'):
        if not isinstance(param, DayOfWeek):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                DayOfWeek, _DAY_OF_WEEK_EXAMPLE))
        return param
    
DayOfWeek.MONDAY = DayOfWeek(0)
DayOfWeek.TUESDAY = DayOfWeek(1)
DayOfWeek.WEDNESDAY = DayOfWeek(2)
DayOfWeek.THURSDAY = DayOfWeek(3)
DayOfWeek.FRIDAY = DayOfWeek(4)
DayOfWeek.SATURDAY = DayOfWeek(5)
DayOfWeek.SUNDAY = DayOfWeek(6)              

_START_OF_MONTH_EXAMPLE = ("StartOfMonth(1) for regular calendar months "
    "starting on the 1st of each month, "
    "StartOfMonth(2) for the \"months\" starting on the 2nd of each month, "
    "etc.")
class StartOfMonth(Immutable):
    __slots__ = ('__dayOfMonth',)
    def __init__(self, dayOfMonth):
        private.checkInt(dayOfMonth, 'dayOfMonth')
        if (dayOfMonth < 1 or dayOfMonth > 28):
            raise ValueError('Invalid dayOfMonth (' +
                str(dayOfMonth) + ') - it cannot be less than 1 '
                'or greater than 28 (to ensure it can work for all '
                'months of all years).')
        self.__dayOfMonth = dayOfMonth
    def _equalityFields(self):
        return (self.__dayOfMonth,)
    @property
    def dayOfMonth(self): return self.__dayOfMonth
    def __repr__(self):
        return 'StartOfMonth(%d)' % self.__dayOfMonth
    def __str__(self):
        return '---%02d' % self.__dayOfMonth
    @classmethod
    def _check(cls, param, paramName='startOfMonth'):
        if not isinstance(param, StartOfMonth):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                StartOfMonth, _START_OF_MONTH_EXAMPLE))
        return param
    
def _minNoDaysInMonth(monthOfYear):
    return [
        0, # dummy value, just so the first month index is 1.
        31, # Jan
        29, # Feb
        31, # Mar
        30, # Apr
        31, # May
        30, # Jun
        31, # Jul
        31, # Aug
        30, # Sep
        31, # Oct
        30, # Nov
        31 # Dec
    ][monthOfYear]

_START_OF_YEAR_EXAMPLE = ("StartOfYear(1, 1) for regular calendar years "
    "starting on the 1st of January, "
    "StartOfYear(4, 6) for \"years\" starting on the 6th of April, etc.")
class StartOfYear(Immutable):
    __slots__ = ('__monthOfYear', '__dayOfMonth')
    def __init__(self, monthOfYear, dayOfMonth):
        private.checkInt(monthOfYear, 'monthOfYear')
        private.checkInt(dayOfMonth, 'dayOfMonth')
        if (monthOfYear < 1 or monthOfYear > 12):
            raise ValueError('Invalid monthOfYear (' + str(monthOfYear) +
                ') - it cannot be less than 1 (January) or greater than 12 '
                '(December).')
        if (dayOfMonth < 1):
            raise ValueError('Invalid dayOfMonth (' + str(dayOfMonth) +
                ') - it cannot be less than 1.');
        if (monthOfYear == 2 and dayOfMonth > 28):
            raise ValueError('Invalid dayOfMonth (' + str(dayOfMonth) + ') - '
                'when when the month is February (2), the day cannot be '
                'greater than 28.')
        noDaysInMonth = _minNoDaysInMonth(monthOfYear)
        if (dayOfMonth > noDaysInMonth):
            raise ValueError('Invalid dayOfMonth (' + str(dayOfMonth) +
                ') - it cannot be greater than ' + str(noDaysInMonth) +
                ' when the month is ' + str(monthOfYear) + '.')
        self.__monthOfYear = monthOfYear
        self.__dayOfMonth = dayOfMonth
    def _equalityFields(self):
        return (self.__monthOfYear, self.__dayOfMonth)
    @property
    def monthOfYear(self): return self.__monthOfYear
    @property
    def dayOfMonth(self): return self.__dayOfMonth
    def __repr__(self):
        return 'StartOfYear(%d, %d)' % (self.__monthOfYear, self.__dayOfMonth)
    def __str__(self):
        return '--%02d-%02d' % (self.monthOfYear, self.dayOfMonth)
    @classmethod
    def _check(cls, param, paramName='startOfYear'):
        if not isinstance(param, StartOfYear):
            raise TypeError(private.wrongSupertypeString(param, paramName,
                StartOfYear, _START_OF_YEAR_EXAMPLE))
        return param
